"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_base app *

:details: lara_django_base app configuration. 
         This provides a generic django app configuration mechanism.
         For more details see:
         https://docs.djangoproject.com/en/4.0/ref/applications/
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>
________________________________________________________________________
"""

import os

from django.apps import AppConfig
from django.conf import settings
from django.urls import include, path

from .object_factory import ObjectFactory


class SyncService(ObjectFactory):
    """SyncService is a factory for SyncBuilderInterface"""

    def get(self, service_id, **kwargs):
        return self.create(service_id.strip().lower(), **kwargs)

from SPARQLWrapper import SPARQLWrapper, JSON, RDF, CSV, TURTLE, DIGEST

from emmopy import get_emmo
from ontopy import World
from rdflib import *


class LaraDjangoBaseConfig(AppConfig):
    name = "lara_django_base"
    url_path = ''
    default_auto_field = "django.db.models.BigAutoField"
    # enter a verbose name for your app: lara_django_base here - this will be used in the admin interface
    verbose_name = "LARA-django Base"
    # lara_app_icon = 'lara_django_base_icon.svg'  # this will be used to display an icon, e.g. in the main LARA menu.
    lara_app_icon = "LARA_logo.svg"

    # sparql = SPARQLWrapper(f'http://{settings.VIRTUOSO_HOST}:{settings.VIRTUOSO_ADMIN_PORT}/sparql-auth/')
    # sparql.setHTTPAuth(DIGEST)
    # sparql.setCredentials(settings.VIRTUOSO_USER, settings.VIRTUOSO_PASSWORD)
    # sparql.setReturnFormat(JSON)
    # def __init__(self, app_name, app_module):
    #     super().__init__(app_name, app_module)

    def ready(self):

        # add urlpatterns 
        from lara_django.urls import urlpatterns

        urlpatterns += [ 
            path(self.url_path, include('lara_django_base.urls', namespace='lara_django_base')),
        ]

        # adding WEBPACK_LOADER

        app_static_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "static", "webpack_bundles")
        stats_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "static", "webpack-stats-base-dev.json")
        settings.WEBPACK_LOADER['BASE'] = {
            "CACHE": not settings.DEBUG,
            "BUNDLE_DIR_NAME": app_static_dir + "/",
            "STATS_FILE": stats_file,
            "POLL_INTERVAL": 0.9,
            "IGNORE": [r".+\.hot-update.js", r".+\.map"],
            "LOADER_CLASS": "webpack_loader.loader.WebpackLoader",
        }

        self.sparql = SPARQLWrapper(f"http://{settings.VIRTUOSO_HOST}:{settings.VIRTUOSO_ADMIN_PORT}/sparql-auth/")
        self.sparql.setHTTPAuth(DIGEST)
        self.sparql.setCredentials(settings.VIRTUOSO_USER, settings.VIRTUOSO_PASSWORD)
        self.sparql.setReturnFormat(JSON)

        # loading the EMMO ontology

        if settings.ONTOLOGY_PATH:
            oso_data_url = os.path.join(settings.ONTOLOGY_PATH, "lara_django_data_corr.ttl") # "oso_data.ttl")
        else:
            # oso_data_filename = "../ontologies/oso_data.ttl"
            oso_data_url = "https://gitlab.com/opensourcelab/scientificdata/ontologies/openscienceontology/data/-/raw/main/ontologies/oso_data.ttl"

        self.lara_onto_world = World()
        self.lara_onto = self.lara_onto_world.get_ontology(oso_data_url)
        self.lara_onto.load()  # reload_if_newer = True
        self.lara_onto.sync_python_names()  # synchronize annotations

        print(f"LARA ontologies loaded ... : {oso_data_url}" )

        self.lara_onto_graph = self.lara_onto_world.as_rdflib_graph()
        lara_onto_iri = "http://lara_onto.info/lara_onto-inferred#"
        abox_iri = "http://w3id.org/lara/lara_abox#"
        tbox_iri = "http://w3id.org/lara/lara__tbox#"
        self.lara_onto_graph.bind("lara_onto", lara_onto_iri)
        self.lara_onto_graph.bind("tbox", tbox_iri)
        self.lara_onto_graph.bind("abox", abox_iri)
        self.lara_onto_graph.bind("owl", "http://www.w3.org/2002/07/owl#")

        self.sync_service = SyncService()

        LARA_welcome_txt = (
            "_____________________________________________\n\n"
            "  Welcome to                                 \n"
            "   __      __    ____    __                  \n"
            "  (  )    /__\  (  _ \  /__\                 \n"
            "   )(__  /(__)\  )   / /(__)\                \n"
            "  (____)(__)(__)(_)\_)(__)(__)               \n"
            "                                             \n"
            "   To access LARA go to:                     \n"
            f"     https://{settings.LARA_HOSTNAME}:{settings.LARA_PORT}\n"
            "        \n"
            "_____________________________________________\n\n"
        )

        print(LARA_welcome_txt)



# -*- coding: utf-8 -*-
"""_____________________________________________________________________

:PROJECT: LARAsuite

*generate unittests *

:details: generate_grpc_client - a unittests generator for LARA-django.
          This is a LARA-django-base management command to generate a unittests for a
          given app. The generated client is a python module that can be used to
          communicate with the gRPC server of the given app.

          requirements:
                python-lorem
:usage: lara-django-dev generate_unittests --author-name "mark doerr" --author-email "mark.doerr@uni-greifswald.de" --force lara_django_base

:authors: mark doerr <mark.doerr@uni-greifswald.de>, Kyle Moynahan <moynahan@campus.tu-berlin.de>

.. note:: -
.. todo:: - https://stackoverflow.com/questions/11380413/python-unittest-passing-arguments
          - speed improvemnts https://medium.com/@thehackadda/tips-for-speeding-up-your-django-tests-35bab8250760
________________________________________________________________________
"""

import re
import os
import logging
import lorem  # pip install python-lorem
import random
import uuid
import json
import pdb
from datetime import datetime

from django.apps import apps
from django.conf import settings
from django.core.management.base import LabelCommand, CommandError
from django.db import models

lara_django_apps = tuple(filter(lambda x: "lara_django" in x, settings.__dict__["INSTALLED_APPS"]))


# Configurable constants
MAX_LINE_WIDTH = getattr(settings, "MAX_LINE_WIDTH", 120)
INDENT_WIDTH = getattr(settings, "INDENT_WIDTH", 4)

UNITTEST_FILE_HEADER = """\"\"\"_____________________________________________________________________

:PROJECT: LARAsuite

*{lara_app} - {model.__name__} gRPC API unittests *

:details: {lara_app} - {model.__name__} gRPC API unittests.
         - generated with 'lara-django-dev generate_unittests_grpc --author-name {lara_author} --author-email {lara_author_email} {lara_app}'
         to run the tests with pytest use:
            pytest --durations=0 -s --num-instances 1 --host localhost --port 50051 [path-to-tests]
:authors: {lara_author} <{lara_author_email}>

.. note:: -
.. todo:: -
________________________________________________________________________
\"\"\"\n"""


def indent(level=1, text="", end="\n"):
    return "\n".join([(level * INDENT_WIDTH * " ") + line for line in text.split("\n")]) + end


def lower_strip(s):
    return s.lower().replace(" ", "").strip()


class Command(LabelCommand):
    help = """Generate a unittests for a given LARA app (models)"""
    args = "[app_name, author_name, author_email, num_instances]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument("app_name", help="Name of the LARA app to generate the unittests for")

        parser.add_argument("--author-name", help="Name of the author", default="lara")
        parser.add_argument("--author-email", help="Email of the author", default="lara@lara.org")
        parser.add_argument("--num-instances", help="Number of test cases to generate", default="1")
        parser.add_argument("--force", action="store_true", help="overwrite existing files", default=False)
        # parser.add_argument('model_name', nargs='*')

    # @signalcommand
    def handle(self, *args, **options):
        self.app_name = options["app_name"]

        logging.warning("!! only a scaffold is generated, please check/add content to the generated files !!!")

        try:
            app = apps.get_app_config(self.app_name)
        except LookupError:
            self.stderr.write("This command requires an existing app name as argument")
            self.stderr.write("Available apps:")
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write("    %s" % label)
            return

        model_res = []
        # for arg in options['model_name']:
        #     model_res.append(re.compile(arg, re.IGNORECASE))

        AppUnittestGenerator(app, model_res, options)

        # self.stdout.write()


class AppUnittestGenerator:
    def __init__(self, app_config, model_res, options):
        self.app_config = app_config
        self.model_res = model_res

        self.app_name = options["app_name"]
        self.author_name = options["author_name"]
        self.author_email = options["author_email"]
        self.num_instances = options["num_instances"]
        self.force_overwrite = options["force"]

        self.model_dict = {}
        self.internal_model_name = set()
        self.external_model_name = set()
        self.external_model_names_all = set()
        
        self.model_names = [model.__name__ for model in self.app_config.get_models()]
        self.foreignkey_fields = {}
        self.required_grpc_imports = ["lara_django_base", self.app_name]
        self.required_list_functions = {}

        for model in self.app_config.get_models():
            self.unittests_str = ""
            self.unittests_header_str = ""
            self.unittests_imports_str = ""
            self.unittests_fixture_str = ""
            self.generate_unittests(model)

        self.generate_unittest_config(model)

    def generate_unittests(self, model):
        self.generate_unittests_header(model)
        self.generate_model_tests(model)
        self.generate_unittests_fixtures(model)
        self.generate_unittests_imports(model)

        self.unittests_str = self.unittests_header_str + self.unittests_imports_str + self.unittests_fixture_str + self.unittests_str

        # save unittests to file
        unittest_filename = f"test_{self.app_name}_{model.__name__}_grpc.py"

        print(f"writing {unittest_filename} to {os.getcwd()}")
        #print(self.unittests_str)

        if os.path.isfile(unittest_filename):
            if self.force_overwrite:
                with open(unittest_filename, "w") as f:
                    f.write(self.unittests_str)
            else:
                append = input(
                    f"{unittest_filename} already exists,\n press 'a' to append,\n press 'o' to overwrite,\n or 'n' to do nothing: "
                )
                if append.lower() == "a":
                    with open(unittest_filename, "a") as f:
                        f.write(self.unittests_str)
                elif append.lower() == "o":
                    with open(unittest_filename, "w") as f:
                        f.write(self.unittests_str)
        else:
            with open(unittest_filename, "w") as f:
                f.write(self.unittests_str)

    def generate_unittests_header(self, model):
        self.unittests_header_str = UNITTEST_FILE_HEADER.format(
            lara_app=self.app_name,
            lara_author=self.author_name,
            lara_author_email=self.author_email,
            lara_num_instances=self.num_instances,
            model=model,
        )

    def generate_unittests_imports(self, model):
        # pdb.set_trace()
        # models = [model for model in self.app_config.get_models()]
        # fields = [field for field in models[3]._meta.get_fields()]
        grpc_imports = ""
        async_functions = ""
        for app in self.required_grpc_imports:
            grpc_imports += f"import {app}_grpc.v1.{app}_pb2 as {app}_pb2\n"
            grpc_imports += f"import {app}_grpc.v1.{app}_pb2_grpc as {app}_pb2_grpc\n\n"
       
        self.unittests_imports_str += (
            f"""
import json
import logging
import uuid
import asyncio
import grpc
from google.protobuf.struct_pb2 import Struct
import pytest
import pytest_asyncio
import random

{grpc_imports}

#import lara_django_APP_grpc.v1.lara_django_APP_pb2 as lara_django_APP_pb2
#import lara_django_APP_grpc.v1.lara_django_APP_pb2_grpc as lara_django_APP_pb2_grpc

\n""" )
        
    def generate_unittests_fixtures(self, model):
        
        self.unittests_fixture_str += f"""@pytest_asyncio.fixture(scope="module")
async def {model.__name__.lower()}_client(grpc_channel):
    return {self.app_name}_pb2_grpc.{model.__name__}ControllerStub(grpc_channel)
\n"""
        
        self.unittests_fixture_str += "# " + ", ".join([f"{external_model.lower()}_ids" for external_model in self.external_model_name]) + "\n\n"
        
    def generate_model_tests(self, model):
        self.unittests_str += f"""
@pytest_asyncio.fixture(scope="module")
async def {model.__name__.lower()}_minimal({model.__name__.lower()}_client, num_instances, namespace_ids):
    # !! add all fixtures in parameter list above !!

    namespace_id_list = namespace_ids

    {model.__name__.lower()}_list = []

    for i in range(num_instances):
        print("Creating  {model.__name__}")

        test_{model.__name__.lower()} = dict(
            # !! add fields here !!
        )
        create_res = await {model.__name__.lower()}_client.Create({self.app_name.lower()}_pb2.{model.__name__}Request(** test_{model.__name__.lower()}))
        {model.__name__.lower()}_list.append(create_res.{model.__name__.lower()}_id)
    
    print({model.__name__.lower()}_list)
        
    yield {model.__name__.lower()}_list

@pytest_asyncio.fixture(scope="module")
async def {model.__name__.lower()}({model.__name__.lower()}_client, num_instances, namespace_ids):

    namespace_id_list = namespace_ids

    {model.__name__.lower()}_list = []

    for i in range(num_instances):
        print("Creating  {model.__name__}")

        test_{model.__name__.lower()} = dict(
{ self.parse_fields_for_init(model) }
        )
        create_res = await {model.__name__.lower()}_client.Create({self.app_name.lower()}_pb2.{model.__name__}Request(** test_{model.__name__.lower()}))
        {model.__name__.lower()}_list.append(create_res.{model.__name__.lower()}_id)

    print({model.__name__.lower()}_list)
        
    yield {model.__name__.lower()}_list

@pytest.mark.asyncio(loop_scope="session")
class Test{model.__name__}MinimalCRUD:
    loop: asyncio.AbstractEventLoop

    @pytest.mark.skip(reason="Not fully implemented yes")
    async def test_{model.__name__.lower()}_create(self, {model.__name__.lower()}_client, {model.__name__.lower()}_minimal, num_instances, namespace_ids):
        print("Creating: {model.__name__.lower()}_minimal")
        print({model.__name__.lower()}_minimal)
        assert len({model.__name__.lower()}_minimal) == num_instances

    @pytest.mark.skip(reason="Not fully implemented yes")
    async def test_{model.__name__.lower()}_retrieve(self, {model.__name__.lower()}_client, {model.__name__.lower()}_minimal):
        print("Retrieving {model.__name__.lower()}_minimal")
        print({model.__name__.lower()}_minimal)
        response = await {model.__name__.lower()}_client.Retrieve({self.app_name.lower()}_pb2.{model.__name__}RetrieveRequest({model.__name__.lower()}_id={model.__name__.lower()}_minimal[0]))
        assert response.{model.__name__.lower()}_id == {model.__name__.lower()}_minimal[0]

    @pytest.mark.skip(reason="Not fully implemented yet")
    async def test_{model.__name__.lower()}_update(self, {model.__name__.lower()}_client, {model.__name__.lower()}_minimal):
        print("Updating {model.__name__.lower()}_minimal")
        print({model.__name__.lower()}_minimal)
        {model.__name__.lower()}_updated = dict({model.__name__.lower()}_id={model.__name__.lower()}_minimal[0], name="test_{model.__name__.lower()}_name_updated", description="{model.__name__.lower()} description updated")

        response_update = await {model.__name__.lower()}_client.Update({self.app_name.lower()}_pb2.{model.__name__}Request(**{model.__name__.lower()}_updated))
        
        response_retr = await {model.__name__.lower()}_client.Retrieve({self.app_name.lower()}_pb2.{model.__name__}RetrieveRequest({model.__name__.lower()}_id={model.__name__.lower()}_minimal[0]))

        assert response_retr.name == "test_{model.__name__.lower()}_name_updated"

    @pytest.mark.skip(reason="Not fully implemented yet")
    async def test_{model.__name__.lower()}_destroy(self, {model.__name__.lower()}_client, {model.__name__.lower()}_minimal):
        print("Deleting all items in {model.__name__.lower()}_minimal")
        print({model.__name__.lower()}_minimal)

        for {model.__name__.lower()}_id in {model.__name__.lower()}_minimal:
            response = await {model.__name__.lower()}_client.Destroy({self.app_name.lower()}_pb2.{model.__name__}DestroyRequest({model.__name__.lower()}_id={model.__name__.lower()}_id))
            print(f"destroying item: {{{model.__name__.lower()}_id}}")
            # assert response == [{model.__name__.lower()}_id]

@pytest.mark.asyncio(loop_scope="session")
class Test{model.__name__}CRUD:
    loop: asyncio.AbstractEventLoop

    @pytest.mark.skip(reason="Not fully implemented yes")
    async def test_{model.__name__.lower()}_create(self, {model.__name__.lower()}_client, {model.__name__.lower()}, num_instances, namespace_ids):
        print("Creating: {model.__name__.lower()}")
        print({model.__name__.lower()})

        assert len({model.__name__.lower()}) == num_instances

    @pytest.mark.skip(reason="Not fully implemented yes")
    async def test_{model.__name__.lower()}_list(self, {model.__name__.lower()}_client, {model.__name__.lower()}):
        print("List {model.__name__.lower()}")
        print({model.__name__.lower()})

        filter_as_dict = {{"name__contains": "test_{model.__name__.lower()}_name"}}
        metadata = (("filters", (json.dumps(filter_as_dict))),)
        response = await {model.__name__.lower()}_client.List({self.app_name.lower()}_pb2.{model.__name__}ListRequest(), metadata=metadata)
        assert len(response.results) == len({model.__name__.lower()}) 

    @pytest.mark.skip(reason="Not fully implemented yes")
    async def test_{model.__name__.lower()}_retrieve(self, {model.__name__.lower()}_client, {model.__name__.lower()}):
        print("Retrieving {model.__name__.lower()}")
        print({model.__name__.lower()})

        for {model.__name__.lower()}_id in {model.__name__.lower()}:
            response = await {model.__name__.lower()}_client.Retrieve({self.app_name.lower()}_pb2.{model.__name__}RetrieveRequest({model.__name__.lower()}_id={model.__name__.lower()}_id))
            assert response.{model.__name__.lower()}_id == {model.__name__.lower()}_id

    @pytest.mark.skip(reason="Not fully implemented yet")
    async def test_{model.__name__.lower()}_update(self, {model.__name__.lower()}_client, {model.__name__.lower()}):
        print("Updating {model.__name__.lower()}")
        print({model.__name__.lower()})

        {model.__name__.lower()}_updated = dict({model.__name__.lower()}_id={model.__name__.lower()}[0], name="test_{model.__name__.lower()}_name_updated", description="{model.__name__.lower()} description updated")

        response_update = await {model.__name__.lower()}_client.Update({self.app_name.lower()}_pb2.{model.__name__}Request(**{model.__name__.lower()}_updated))
        
        response_retr = await {model.__name__.lower()}_client.Retrieve({self.app_name.lower()}_pb2.{model.__name__}RetrieveRequest({model.__name__.lower()}_id={model.__name__.lower()}[0]))

        assert response_retr.name == "test_{model.__name__.lower()}_name_updated"

    @pytest.mark.skip(reason="Not fully implemented yet")
    async def test_{model.__name__.lower()}_destroy(self, {model.__name__.lower()}_client, {model.__name__.lower()}):
        print("Destroy all items in {model.__name__.lower()}")
        print({model.__name__.lower()})
        
        for {model.__name__.lower()}_id in {model.__name__.lower()}:
            response = await {model.__name__.lower()}_client.Destroy({self.app_name.lower()}_pb2.{model.__name__}DestroyRequest({model.__name__.lower()}_id={model.__name__.lower()}_id))
            print(f"destroying item: {{{model.__name__.lower()}_id}}")
            # assert response == [{model.__name__.lower()}_id]
    

"""

    def parse_fields_for_init(self, model):
        # TODO: add automatic handling here
        fields_str = ""

        fk_fields = {}

        for field in model._meta.get_fields():
            #print(field)
            unique_id = ""
            # print(f"field: {field} --> {type(field)} \n")

            # if model.__name__.lower() == "sample":
            #     pdb.set_trace()
            if (
                isinstance(field, models.fields.related.ManyToManyField)
                or isinstance(field, models.fields.reverse_related.ManyToOneRel)
                or isinstance(field, models.fields.reverse_related.ManyToManyRel)
                or isinstance(field, models.fields.reverse_related.OneToOneRel)
                or isinstance(field, models.fields.reverse_related.ManyToManyRel)
            ):
                required = False
                
            else:
                required = field.blank == False
            if required:
                unique_id = "_{i}"

            match field:
                case models.fields.related.ManyToOneRel():
                    pass
                case models.fields.UUIDField():
                    # TODO: need to handle these
                    # fields_str += indent(
                    #     3,
                    #     f"{field.model.__name__.lower()}_id = {field.model.__name__.lower().replace('_','')}_ids[random.randrange(len({field.model.__name__.lower().replace('_','')}_ids))],  # required:{required} (ManyToManyField -> {field.model.__name__}) {field.help_text}3",
                    # )

                    fk_fields[field.model.__name__.lower()] = {
                        "related_model": field.model.__name__.replace("_", ""),
                        "app": None,
                    }
                    # pdb.set_trace()
                    # self.external_model_name.add("UUID")
                    # pdb.set_trace()

                case models.fields.BooleanField():
                    fields_str += indent(
                        3,
                        f"{field.name} = {random.choice(('True', 'False') ) },   # required:{required} (bool) {field.help_text}",
                    )
                case models.fields.URLField():
                    url_str = f"http://www.{lorem.get_word()}.org/lara"
                    if field.name == "url":
                        url_str = f"http://www.{lorem.get_word()}.org/lara"
                    elif field.name == "iri":
                        url_str = f"http://w3id.org/lara/{model.__name__}_{lorem.get_word()}_{{i}}"

                    fields_str += indent(
                        3,
                        f'{field.name} = f"{url_str}",   # required:{required} (URL - str) {field.help_text}',
                    )
                case models.fields.CharField():
                    
                    fields_str += indent(
                        3,
                        f'{field.name} = f"{lorem.get_word()}{unique_id}",   # required:{required} (str) {field.help_text}',
                    )

                case models.fields.TextField():
                    if field.name == "name":
                        fields_str += indent(
                            3,
                            f'{field.name} = f"test_{model.__name__.lower()}_name_{{i}}",   #required:{required} (str) {field.help_text}',
                        )
                    elif field.name == "name_full":
                        fields_str += indent(
                            3,
                            f'{field.name} = f"test_{model.__name__.lower()}_name_full_{{i}}",   #required:{required} (str) {field.help_text}',
                        )
                    elif field.name == "title":
                        fields_str += indent(
                            3,
                            f'{field.name} = f"test_{model.__name__.lower()}_title_{{i}}",   #required:{required} (str) {field.help_text}',
                        )
                    else:
                        fields_str += indent(
                            3,
                            f'{field.name} = f"{lorem.get_sentence()}{unique_id}",   # required:{required}(str) {field.help_text}',
                        )

                case models.fields.IntegerField():
                    fields_str += indent(
                        3,
                        f"{field.name} = {random.randint(1,100)},   #required:{required} (int) {field.help_text}",
                    )

                case models.fields.FloatField():
                    fields_str += indent(
                        3,
                        f"{field.name} = {random.uniform(1.0, 100.0)},   #required:{required} (float) {field.help_text}",
                    )

                case models.fields.DecimalField():
                    fields_str += indent(
                        3,
                        f"{field.name} = {random.uniform(1.0, 100.0)},   #required:{required} (float) {field.help_text}",
                    )

                case models.fields.DateTimeField():
                    fields_str += indent(
                        3,
                        f'{field.name} = "{datetime.now()}",   #required:{required} (datetime) {field.help_text}',
                    )

                case models.fields.DateField():
                    fields_str += indent(
                        3,
                        f'{field.name} = "{datetime.now().date()}",   #required:{required} (date) {field.help_text}',
                    )

                case models.fields.json.JSONField():
                    # random json
                    rand_json = json.dumps(
                        {"fields": {"key1": lorem.get_word(), "key2": lorem.get_word(), "key3": lorem.get_word()}}
                    )
                    fields_str += indent(
                        3,
                        f"{field.name} = Struct().update({rand_json}),   #required:{required} (str) {field.help_text}",
                    )

                case models.fields.files.FileField():
                    fields_str += indent(2, f"### FILES NOT HANDLED YET")
                    fields_str += indent(
                        3, f'{field.name} = "my_file.md",  #required:{required} (file) {field.help_text}'
                    )

                case models.fields.related.ForeignKey():
                    # print(f"{field.name} --- attributes: {field.__dict__}")
                    fields_str += indent(
                        3,
                        f"{field.name} = random.choice({field.related_model.__name__.lower()}_ids) if len({field.related_model.__name__.lower()}_ids) > 0 else None,  # required:{required} (ForeignKey ->  {field.related_model.__name__.lower()}) {field.help_text}",
                    )
                    fk_fields[field.name] = {"related_model": field.related_model.__name__, "app": None}
                    
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

                case models.OneToOneField():
                    fields_str += indent(
                        3,
                        f"{field.name} =  ,  # required:{required} (OneToOneField -> {field.related_model.__name__}) {field.help_text}",
                    )
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

                case models.fields.related.ManyToManyField():
                    # print(f"field attributes: {field.__dict__}")
                    fields_str += indent(
                        3,
                        f"{field.name} = [random.choice({field.related_model.__name__.lower()}_ids)] if len({field.related_model.__name__.lower()}_ids) > 0 else None,  # required:{required} (ManyToManyField -> {field.related_model.__name__}) {field.help_text}",
                    )
                    self.external_model_name.add(field.related_model.__name__)

                    fk_fields[field.name] = {"related_model": field.related_model.__name__, "app": None}
                    self.model_dict[field.related_model.__name__] = field.related_model

        self.external_model_names_all.update(self.external_model_name)
        return fields_str

    def get_relational_fields(self, model):
        foreignkey_fields = set()
        related_names = set()
        m2m_fields = set()
        one2one_fields = set()

        for field in model._meta.get_fields():
            print(f"field: {field} --> {type(field)} ")
            if isinstance(field, models.ForeignKey):
                foreignkey_fields.add(field.name)
            if isinstance(field, models.fields.related.ForeignKey):
                pass
                # print(f"field attributes: {field.__dict__}")
                # related_names.add(field.name)
            if isinstance(field, models.fields.related.ManyToManyField):
                # print(f"field attributes: {field.__dict__}")
                m2m_fields.add(field.name)
            if isinstance(field, models.fields.reverse_related.ManyToManyRel):
                # print(f"field attributes: {field.__dict__}")
                related_names.add(field.name)
                related_names.add(field.related_name)
                related_names.add(field.related_query_name)
                related_names.add(field.field.name)
                # related_names.add(field.field.related_name)
                related_names.add(field.field.related_query_name)
            if isinstance(field, models.fields.related.OneToOneField):
                one2one_fields.add(field.name)
            if isinstance(field, models.fields.reverse_related.OneToOneRel):
                related_names.add(field.name)
                related_names.add(field.related_name)
                related_names.add(field.related_query_name)
        return [foreignkey_fields, m2m_fields, one2one_fields, related_names]

    def generate_unittest_config(self, model):
        self.unittests_config_str = ""
        self.unittests_header_str = ""
        model.__name__= "PyTest Config"

        self.generate_unittests_header(model)
        self.generate_unittests_config_body()
        self.generate_external_fixtures()

        self.unittests_conf_str = self.unittests_header_str + self.unittests_config_str

        self.write_pytest_config()

    def generate_unittests_config_body(self):
        grpc_imports = ""
        for app in self.required_grpc_imports:
            grpc_imports += f"import {app}_grpc.v1.{app}_pb2 as {app}_pb2\n"
            grpc_imports += f"import {app}_grpc.v1.{app}_pb2_grpc as {app}_pb2_grpc\n\n"

        self.unittests_config_str += f"""        
import grpc
import asyncio
import pytest
import pytest_asyncio

{grpc_imports}

def pytest_addoption(parser):
    parser.addoption("--host", action="store", default="localhost")
    parser.addoption("--port", action="store", default=50066)
    parser.addoption("--num-instances", action="store", default=1, help="Number of instances to create test")


@pytest.fixture(scope="session")
def num_instances(pytestconfig):
    return int(pytestconfig.getoption("num_instances"))


@pytest_asyncio.fixture(scope="session")
async def grpc_channel(pytestconfig):
    print(f"Creating gRPC channel: {{pytestconfig.getoption('host')}}:{{pytestconfig.getoption('port')}}" )
    channel = grpc.aio.insecure_channel(f"{{pytestconfig.getoption('host')}}:{{pytestconfig.getoption('port')}}")

    yield channel

    print(f"\\nAll tests executed !\\nClosing gRPC channel: {{pytestconfig.getoption('host')}}:{{pytestconfig.getoption('port')}}" )
    await channel.close()

"""

    def generate_external_fixtures(self):
        for external_model in self.external_model_names_all:
            self.unittests_config_str  += f"""@pytest_asyncio.fixture(scope="module")
async def {external_model.lower()}_ids(pytestconfig, grpc_channel):
    created_item_ids = []
    {external_model.lower()}_client = lara_django_APP_pb2_grpc.{external_model}ControllerStub(grpc_channel)
    res = await {external_model.lower()}_client.List(lara_django_APP_pb2.{external_model}ListRequest())

    {external_model.lower().replace('_','')}_ids = [r.{external_model.lower().replace('_','')}_id for r in res.results]

    if len(res.results) == 0:
        for i in range(3):
            print(f"Creating  {external_model.lower().replace('_','')}")
            test_{external_model.lower().replace('_','')} = dict(
                # !! add fields here !!
            )
            create_res = await {external_model.lower()}_client.Create(lara_django_APP_pb2.{external_model}Request(** test_{external_model.lower().replace('_','')}))
            created_item_ids.append(create_res.{external_model.lower().replace('_','')}_id)
        {external_model.lower().replace('_','')}_ids = created_item_ids

    yield {external_model.lower().replace('_','')}_ids

    if len(created_item_ids) > 0:
        for {external_model.lower().replace('_','')}_id in created_item_ids:
            response = await {external_model.lower()}_client.Destroy(lara_django_APP_pb2.{external_model}DestroyRequest({external_model.lower().replace('_','')}_id={external_model.lower().replace('_','')}_id))
            print(f"\\n+++ finally destroying external {external_model.lower().replace('_','')}_id: {{{external_model.lower().replace('_','')}_id}}")
\n"""

    def write_pytest_config(self):
        # save unittests to file
        unittest_config_filename = "conftest.py"

        print(f"writing {unittest_config_filename} to {os.getcwd()}")
        #print(self.unittests_str)

        if os.path.isfile(unittest_config_filename):
            if self.force_overwrite:
                with open(unittest_config_filename, "w") as f:
                    f.write(self.unittests_conf_str)
            else:
                append = input(
                    f"{unittest_config_filename} already exists,\n press 'a' to append,\n press 'o' to overwrite,\n or 'n' to do nothing: "
                )
                if append.lower() == "a":
                    with open(unittest_config_filename, "a") as f:
                        f.write(self.unittests_conf_str)
                elif append.lower() == "o":
                    with open(unittest_config_filename, "w") as f:
                        f.write(self.unittests_conf_str)
        else:
            with open(unittest_config_filename, "w") as f:
                f.write(self.unittests_conf_str)
        




"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_myproj gRPC services*

:details: lara_django_myproj gRPC services.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

## generated with django-socio-grpc generateprpcinterface lara_django_base  (LARA-version)

import asyncio

import lara_django_base_grpc.v1.lara_django_base_pb2 as lara_django_base_pb2
from django_socio_grpc import generics, mixins
from django_socio_grpc.decorators import grpc_action

from lara_django_base.grpc.mixins import (
    AsyncDownloadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncUploadModelMixin,
)
from lara_django_base.models import (
    Address,
    City,
    Country,
    CountrySubdivision,
    Currency,
    DataType,
    ExternalResource,
    ExtraData,
    GeoLocation,
    ItemStatus,
    LARASyncServer,
    Location,
    MediaType,
    Namespace,
    PhysicalStateMatter,
    PhysicalUnit,
    Room,
    RoomCategory,
    ScientificConstant,
    SubdivisionType,
    Tag,
)

from ..filters import (
    AddressFilterSet,
    CityFilterSet,
    CountryFilterSet,
    CountrySubdivisionFilterSet,
    CurrencyFilterSet,
    DataTypeFilterSet,
    ExternalResourceFilterSet,
    ExtraDataFilterSet,
    GeoLocationFilterSet,
    ItemStatusFilterSet,
    LARASyncServerFilterSet,
    LocationFilterSet,
    MediaTypeFilterSet,
    NamespaceFilterSet,
    PhysicalStateMatterFilterSet,
    PhysicalUnitFilterSet,
    RoomCategoryFilterSet,
    RoomFilterSet,
    ScientificConstantFilterSet,
    SubdivisionTypeFilterSet,
    TagFilterSet,
)
from .serializers import (
    AddressProtoSerializer,
    CityProtoSerializer,
    CountryProtoSerializer,
    CountrySubdivisionProtoSerializer,
    CurrencyProtoSerializer,
    DataTypeProtoSerializer,
    ExternalResourceProtoSerializer,
    ExtraDataProtoSerializer,
    GeoLocationProtoSerializer,
    ItemStatusProtoSerializer,
    LARASyncServerProtoSerializer,
    LocationProtoSerializer,
    MediaTypeProtoSerializer,
    NamespaceByURIProtoSerializer,
    NamespaceProtoSerializer,
    PhysicalStateMatterProtoSerializer,
    PhysicalUnitProtoSerializer,
    RoomCategoryProtoSerializer,
    RoomProtoSerializer,
    ScientificConstantProtoSerializer,
    SubdivisionTypeProtoSerializer,
    TagProtoSerializer,
)


class DataTypeService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = DataType.objects.all()
    serializer_class = DataTypeProtoSerializer
    filterset_class = DataTypeFilterSet


class MediaTypeService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = MediaType.objects.all()
    filterset_class = MediaTypeFilterSet
    serializer_class = MediaTypeProtoSerializer


class NamespaceService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Namespace.objects.all()
    filterset_class = NamespaceFilterSet
    serializer_class = NamespaceProtoSerializer

    lookup_field = "namespace_id"


class NamespaceByURIService(
    generics.GenericService, mixins.AsyncListModelMixin, mixins.AsyncRetrieveModelMixin
):
    serializer_class = NamespaceByURIProtoSerializer
    queryset = Namespace.objects.all().order_by("uri")

    lookup_field = "uri"


class TagService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Tag.objects.all()
    serializer_class = TagProtoSerializer
    filterset_class = TagFilterSet


class ItemStatusService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = ItemStatus.objects.all()
    serializer_class = ItemStatusProtoSerializer
    filterset_class = ItemStatusFilterSet


class ExtraDataService(
    generics.AsyncModelService,
    mixins.AsyncStreamModelMixin,
    AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,
):
    pb2 = lara_django_base_pb2
    queryset = ExtraData.objects.all()
    serializer_class = ExtraDataProtoSerializer
    filterset_class = ExtraDataFilterSet


class PhysicalUnitService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = PhysicalUnit.objects.all()
    serializer_class = PhysicalUnitProtoSerializer
    filterset_class = PhysicalUnitFilterSet


class ExternalResourceService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = ExternalResource.objects.all()
    serializer_class = ExternalResourceProtoSerializer
    filterset_class = ExternalResourceFilterSet


class PhysicalStateMatterService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = PhysicalStateMatter.objects.all()
    serializer_class = PhysicalStateMatterProtoSerializer
    filterset_class = PhysicalStateMatterFilterSet


class ScientificConstantService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = ScientificConstant.objects.all()
    serializer_class = ScientificConstantProtoSerializer
    filterset_class = ScientificConstantFilterSet


class CurrencyService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = Currency.objects.all()
    serializer_class = CurrencyProtoSerializer
    filterset_class = CurrencyFilterSet


class GeoLocationService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = GeoLocation.objects.all()
    serializer_class = GeoLocationProtoSerializer
    filterset_class = GeoLocationFilterSet


class CountryService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = Country.objects.all()
    serializer_class = CountryProtoSerializer
    filterset_class = CountryFilterSet


class SubdivisionTypeService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = SubdivisionType.objects.all()
    serializer_class = SubdivisionTypeProtoSerializer
    filterset_class = SubdivisionTypeFilterSet


class CountrySubdivisionService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = CountrySubdivision.objects.all()
    serializer_class = CountrySubdivisionProtoSerializer
    filterset_class = CountrySubdivisionFilterSet


class CityService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = City.objects.all()
    serializer_class = CityProtoSerializer
    filterset_class = CityFilterSet


class AddressService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Address.objects.all()
    serializer_class = AddressProtoSerializer
    filterset_class = AddressFilterSet


class RoomCategoryService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = RoomCategory.objects.all()
    serializer_class = RoomCategoryProtoSerializer
    filterset_class = RoomCategoryFilterSet


class RoomService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = Room.objects.all()
    serializer_class = RoomProtoSerializer
    filterset_class = RoomFilterSet


class LocationService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = Location.objects.all()
    serializer_class = LocationProtoSerializer
    filterset_class = LocationFilterSet


class LARASyncServerService(
    generics.AsyncModelService,
    AsyncRetrieveByNameModelMixin,
    mixins.AsyncStreamModelMixin,
):
    queryset = LARASyncServer.objects.all()
    serializer_class = LARASyncServerProtoSerializer
    filterset_class = LARASyncServerFilterSet



"""_____________________________________________________________________

:PROJECT: lara-django

*lara-django-database maintainance*

:details: lara-django-database maintainance


:authors: mark doerr (mark@uni-greifswald.de)

:date: (creation)          20220927

.. note:: -
.. todo:: -
________________________________________________________________________

**Copyright**:
  This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
  INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  For further Information see LICENSE file that comes with this distribution.
________________________________________________________________________
"""

__version__ = "0.0.1"

import os
import logging
from pathlib import Path


from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command
from django.core.exceptions import ImproperlyConfigured
from django.core.management.color import color_style, no_style
from django.db import DEFAULT_DB_ALIAS, connections
from django.conf import settings

from ..simpleinstall import query_yes_no


class Command(BaseCommand):
    """see https://docs.djangoproject.com/en/1.9/howto/custom-management-commands/ for more details
       using now new argparse mechanism of django > 1.8
    """
    logging.basicConfig(format='%(levelname)s| %(module)s.%(funcName)s:%(message)s',
                        level=logging.DEBUG)  # level=logging.ERROR

    help = 'DJANGO database maintainace.'

    def add_arguments(self, parser):
        """ command line arguments s. https://docs.python.org/2/library/argparse.html#module-argparse """

        # defining named commandline arguments
        parser.add_argument('--delete-database',
                            action='store_true',
                            help='delete DJANGO database')
        parser.add_argument('--migrations',
                            action='store_true',
                            help='make all migrations')
        parser.add_argument('--delete-migrations',
                            action='store_true',
                            help='delete all migrations')

        parser.add_argument(
            '--force', action='store_true',
            help='Force execution of the initialization process.',
        )

    def handle(self, *args, **options):
        """Dispatcher based on commandline arguments/options"""

        if options['migrations']:
            self.make_all_migrations()

        if options['delete_migrations']:
            self.delete_all_migrations()

        if options['delete_database']:
            self.delete_db()

    def make_all_migrations(self):

        call_command('wait_for_db')

        call_command('makemigrations')

        if self.num_migrations() > 0:
            logging.debug("unresolved migrations")
            self.make_app_migrations()

        call_command('migrate')

    def delete_all_migrations(self, force=False):
        # TODO: not working yet
        for app in settings.LARA_APPS:
            try:
                #logging.debug("migrating : {}".format(app))
                # os.remove(settings.)
                migration_dir = os.path.join(
                    settings.BASE_DIR, '..', '..', app, 'migrations/*')
                print("remove migrations of app", app,
                      " in migr. dir:",  migration_dir)
            except FileNotFoundError as err:
                logging.error(
                    "Migrations dir {0} not found ! [{1}]".format(migration_dir, err))

    def delete_db(self, force=False):
        "removing database"
        logging.info("rather use:\n lara-django-dev reset_db")

        if force:
            del_db = True
        else:
            db_filename = settings.DATABASES['default']['NAME']

            del_db = query_yes_no("Shall I REALLY delete the LARA database\n{}\nand re-initialise it ?".format(db_filename),
                                  default_answer="no", help="This removes the sqlite database and fills it with fresh data")
        if del_db:
            #~ run("pwd; echo deleting db...")

            try:
                logging.debug("deleting db [{}]...".format(db_filename))
                # os.remove(db_filename)
                #~ call_command('flush')
                #~ call_command('migrate')
                # try flush instead !! https://docs.djangoproject.com/en/2.1/ref/django-admin/#django-admin-dumpdata
            except FileNotFoundError as err:
                logging.error(
                    "Database {0} not found ! [{1}]".format(db_filename, err))

    def num_migrations(self):
        """return number of migrations
           :param options:
        """
        from django.db.migrations.executor import MigrationExecutor
        try:
            executor = MigrationExecutor(connections[DEFAULT_DB_ALIAS])
        except ImproperlyConfigured:
            # No databases are configured (or the dummy one)
            return

        plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
        if plan:
            return len(plan)
        else:
            return 0

    def make_app_migrations(self, options=None):
        """make individual app migrations
           s. https://docs.djangoproject.com/en/2.0/ref/django-admin/
           :param options:
        """
        #db_filename = settings.DATABASES['default']['NAME']
        for app in settings.LARA_APPS:
            logging.debug("migrating : {}".format(app))
            call_command('makemigrations', app)

    def load_fixtures(self, options=None):
        """load fixtures the LARA database
           :param options:
        """
        for fixture in settings.FIXTURES:
            call_command('loaddata', fixture)



from importlib.metadata import version as get_version

__version__ = get_version(__package__)

__version_info__ = tuple(
    int(num) if num.isdigit() else num
    for num in __version__.replace("-", ".", 1).split(".")
)

# Re-export public API
from .views_mixins import HtmxResponseMixin

__all__ = [
    "HtmxResponseMixin",
    "__version__",
    "__version_info__",
]



import datetime
import uuid

from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .models_base import (
    ItemStatus,
    Namespace,
    Tag,
)
from .models_location import GeoLocation, Location


def _get_syncserver_types():
    """helper function to get the available LARA server types for synchronisation for choices
    which requires a function to be called"""
    return settings.LARA_SERVER_TYPES


class LARASyncServer(models.Model):
    """LARA synchronisation Server description,
    e.g., for synchronisation between two LARA instances or other server types."""

    class SyncServerTypeChoices(models.TextChoices):
        SYNC_FULL = "f", _("full")
        SYNC_INCREMENTAL = "i", _("incremental")

    model_curi = "lara:base/LARASyncServer"

    laraserver_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(
        blank=True, help_text="human readable name of the Sync server"
    )
    url = models.URLField(
        blank=True,
        help_text="Universal Resource Locator - URL of the LARA server",
    )
    port = models.IntegerField(
        blank=True, null=True, help_text="port of the LARA server"
    )
    server_type = models.CharField(
        max_length=4,
        choices=_get_syncserver_types,
        default="LARA",
        help_text="type of server to be synchronised with",
    )
    connection_json = models.JSONField(
        blank=True,
        null=True,
        help_text="JSON field for more advanced connection information,"
        "e.g., advanced routing, ports, ..., could be even encrypted.",
    )
    auth_json = models.JSONField(
        blank=True,
        null=True,
        help_text="JSON field for authentication information, e.g. username, "
        "password, token, ..., could be even encrypted",
    )
    cert_pub = models.TextField(blank=True, help_text="server public certificate")

    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_tags_related",
        related_query_name="%(app_label)s_%(class)s_tags_related_query",
        blank=True,
        help_text="tags LARA server",
    )
    description = models.TextField(blank=True, help_text="server description")

    class Meta:
        db_table_comment = "LARA Server description, e.g., for synchronisation between two LARA instances"

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    def save(self, *args, **kwargs):
        if self.name is None:
            self.name = self.url

        self.name = self.name.replace(" ", "_").lower().strip()

        super().save(*args, **kwargs)


class SyncStatusAbstr(models.Model):
    """An abstract class to store sync status information, e.g. for synchronisation between two LARA instances"""

    syncstatus_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    status = models.ForeignKey(
        ItemStatus,
        related_name="%(app_label)s_%(class)s_statuses_related",
        related_query_name="%(app_label)s_%(class)s_statuses_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="scheduled, in-progress, interrupted, completed",
    )
    server = models.ForeignKey(
        LARASyncServer,
        related_name="%(app_label)s_%(class)s_servers_related",
        related_query_name="%(app_label)s_%(class)s_servers_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    datetime_last_synced = models.DateTimeField(
        default=timezone.now, help_text="date and time when data was last modified"
    )
    # sync_conflicts = models.FileField(help_text="File containing detailed information
    # about conflicts during synchronisation")

    class Meta:
        abstract = True
        verbose_name_plural = "SyncStatus"


class SynchronisationAbstr(models.Model):
    """Abstract Synchronisation settings class for (recursive) LARA projects synchronisation with other LARA systems

    :param models: _description_
    :type models: _type_
    """

    class SyncModeChoices(models.TextChoices):
        SYNC_FULL = "f", _("full")
        SYNC_INCREMENTAL = "i", _("incremental")

    class SyncPolicyChoices(models.TextChoices):
        UPDATE_SOURCE = "s", _("source")
        UPDATE_TARGET = "t", _("target")
        UPDATE_MERGE = "m", _("merge")

    sync_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of synchronisation",
    )
    name = models.TextField(
        blank=True, unique=True, help_text="name of the synchronisaton"
    )
    server_target = models.ForeignKey(
        LARASyncServer,
        related_name="%(app_label)s_%(class)s_server_related",
        related_query_name="%(app_label)s_%(class)s_server_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="LARA server to be synchronised with",
    )
    datetime_start = models.DateTimeField(
        default=timezone.now, help_text="synchronisation start"
    )
    datetime_end = models.DateTimeField(
        default=datetime.datetime(2066, 12, 31, tzinfo=datetime.timezone.utc),
        help_text="synchronisation end",
    )
    mode = models.CharField(
        max_length=1,
        choices=SyncModeChoices.choices,
        default=SyncModeChoices.SYNC_FULL,
        help_text="synchronisation mode: full, incremental",
    )
    update_policy = models.CharField(
        max_length=1,
        choices=SyncPolicyChoices.choices,
        default=SyncPolicyChoices.UPDATE_MERGE,
        help_text="update policy: source, target, merge",
    )
    event_repeat = models.TextField(
        blank=True,
        help_text="prefect / apache airflow ('@daily', '@weekly') or cron expression"
        "like entry: '12 23 * * *' :  every day at 23:12h do this task",
    )
    event_repeat_json = models.JSONField(
        blank=True,
        null=True,
        help_text="advanced repeat expressions possible, e.g. for uneven occurrences",
    )
    config_json = models.JSONField(
        blank=True,
        null=True,
        help_text="advanced configuration settings in JSON format,"
        "e.g., for advanced settings like filters, mappings, ...",
    )
    logs = models.JSONField(
        null=True, blank=True, help_text="logs of the synchronisation"
    )
    status = models.ForeignKey(
        ItemStatus,
        related_name="%(app_label)s_%(class)s_status_related",
        related_query_name="%(app_label)s_%(class)s_status_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="synchronisation status classifier: planned, started, in_progress, finished",
    )
    log_file = models.FileField(
        upload_to="projects/sync",
        blank=True,
        null=True,
        help_text="log file with rel. path/filename",
    )
    description = models.TextField(
        blank=True, help_text="description of extra data type"
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.name or ""

    def __repr__(self) -> str:
        return self.name or ""

    # duration will be calculated


class CalendarAbstr(models.Model):
    """_summary_
    see https://www.ietf.org/rfc/rfc2445.txt
        RFC 5545

    see also: https://pypi.org/project/ics/
         and  https://pypi.org/project/icalendar/

    :param models: _description_
    :type models: _type_
    """

    name_display = models.TextField(help_text="title of the calendar entry")
    calendar_url = models.URLField(
        blank=True,
        help_text="Universal Resource Locator (URL), to google calendar, iCalendar, .... ",
    )
    summary = models.TextField(
        blank=True, help_text="short summary text of the calendar entry"
    )
    description = models.TextField(blank=True, help_text="text of the calendar entry")
    color = models.TextField(
        # default=ColorBlind16.full_color_scheme[randint(0, 16)].get_web(),
        blank=True,
        help_text="text of the calendar entry",
    )
    # organizer, cn, attendees
    # user - needs to be defined in the instantiation
    ics = models.TextField(
        blank=True, help_text="iCalendar entry in (RFC 5545) file format"
    )
    datetime_start = models.DateTimeField(
        default=timezone.now,
        help_text="start date & time of the event. RFC 5545 - DTSTART",
    )
    datetime_end = models.DateTimeField(
        default=timezone.now, help_text="end datetime of the event. RFC 5545 - DTEND"
    )
    duration = models.TextField(
        blank=True,
        help_text=""" This value type is used to identify properties that contain
         a duration of time. RFC 5545 - DURATION""",
    )
    all_day = models.BooleanField(default=False, help_text="All day event ? (boolean)")

    created = models.DateTimeField(
        default=timezone.now, help_text="end datetime of the event"
    )
    datetime_last_modified = models.DateTimeField(
        default=timezone.now,
        help_text="date and time when calendar entry was last modified",
    )
    timestamp = models.DateTimeField(
        default=timezone.now, help_text="timestamp of the event"
    )
    # transparency
    location = models.ForeignKey(
        Location,
        related_name="%(app_label)s_%(class)s_location_related",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="location",
    )
    geolocation = models.ForeignKey(
        GeoLocation,
        related_name="%(app_label)s_%(class)s_geolocation_related",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="GEO location",
    )

    conference_url = models.URLField(
        blank=True,
        help_text="audio/video conference/meeting Universal Resource Locator (URL), e.g. zoom, jitsi, meet, ...",
    )

    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_tags_related",
        related_query_name="%(app_label)s_%(class)s_tags_related_query",
        blank=True,
        help_text="tags of calendar entry",
    )
    # many parameters / fields of the RFC 5545 should be stored in the ics field
    # participation status
    range = models.TextField(
        blank=True,
        help_text="""To specify the effective range of recurrence instances from
        the instance specified by the recurrence identifier specified by
        the property. RFC 5545 - RANGE """,
    )
    related = models.TextField(
        blank=True,
        help_text="""To specify the relationship of the alarm trigger with
        respect to the start or end of the calendar component. RFC 5545 - RELATED""",
    )
    role = models.TextField(
        blank=True,
        help_text="""To specify the participation role for the calendar user
        specified by the property. RFC 5545 - ROLE""",
    )
    tzid = models.TextField(
        blank=True,
        help_text="""To specify the identifier for the time zone definition for
        a time component in the property value.
        RFC 5545 - TZID""",
    )
    offset_utc = models.IntegerField(
        blank=True,
        default=0,
        help_text=""" This value type is used to identify properties that contain
      an offset from UTC to local time (+ or - num. hours) . RFC 5545 -  UTC-OFFSET""",
    )
    # alarm_method - mail ...
    alarm_repeat_json = models.JSONField(
        blank=True,
        null=True,
        help_text="advanced alarm repeat expressions possible, e.g. for uneven occurrences",
    )
    event_repeat = models.TextField(
        blank=True,
        help_text="apache airflow ('@daily', '@weekly') or cron expression like entry: '12 23 * * *' : "
        "every day at 23:12h do this task",
    )
    event_repeat_json = models.JSONField(
        blank=True,
        null=True,
        help_text="advanced event repeat expressions possible, e.g. for uneven occurrences",
    )
    search_vector = SearchVectorField(null=True)

    class Meta:
        abstract = True
        indexes = (GinIndex(fields=["search_vector"]),)

    # autogenerate ICS on save, if ics is empty.
    # see: https://icalendar.readthedocs.io
    # def save(self, *args, **kwargs):
    #     if self.entity.email == "":
    #         self.entity.email = self.user.email

    #     if self.user.first_name == "":
    #         self.user.first_name = self.entity.name_first

    #     if self.user.last_name == "":
    #         self.user.last_name = self.entity.name_last

    #     if self.user.email == "":
    #         self.user.email = self.entity.email

    #     self.user.save()

    #     super().save(*args, **kwargs)

    def __str__(self):
        return self.name_display or ""

    def __repr__(self):
        return self.name_display or ""



# -*- coding: utf-8 -*-
"""_____________________________________________________________________

:PROJECT: LARAsuite

*generate gRPC client *

:details: generate_grpc_client - a gRPC client generator for LARA-django.
          This is a LARA-django-base management command to generate a gRPC client for a
          given app. The generated client is a python module that can be used to
          communicate with the gRPC server of the given app.

 - `fields` for all local fields

:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

import re
import os
import logging

from django.apps import apps
from django import models
from django.conf import settings
from django.core.management.base import LabelCommand, CommandError
from django.db import models


# Configurable constants
MAX_LINE_WIDTH = getattr(settings, 'MAX_LINE_WIDTH', 120)
INDENT_WIDTH = getattr(settings, 'INDENT_WIDTH', 4)

CLIENT_FILE_HEADER = """""""""_____________________________________________________________________

:PROJECT: LARAsuite

*{lara_app} gRPC client *

:details: {lara_app} client.
         - generated with 'lara-django-dev generate_grpc_client {lara_app}
:authors: {lara_author} <{lara_author_email}>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""""""""



class Command(LabelCommand):
    help = '''Generate a gRPC client for a given LARA app (models)'''
    args = "[app_name, author_name, author_email]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument('app_name', help='Name of the LARA app to generate the gRPC interface for')
            
        parser.add_argument('--author-name', action="store_true", help='Name of the author',  default='lara')
        parser.add_argument('--author-email', action="store_true", help='Email of the author', default='lara@lara.org')
        #parser.add_argument('model_name', nargs='*')

    #@signalcommand
    def handle(self, *args, **options):
        self.app_name = options['app_name']

        logging.warning("!! only a scaffold is generated, please check/add content to the generated files !!!")

        try:
            app = apps.get_app_config(self.app_name)
        except LookupError:
            self.stderr.write('This command requires an existing app name as argument')
            self.stderr.write('Available apps:')
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write('    %s' % label)
            return

        model_res = []
        # for arg in options['model_name']:
        #     model_res.append(re.compile(arg, re.IGNORECASE))

        GRPC_ClientApp(app, model_res, options)

        #self.stdout.write()

class GRPC_ClientApp():
    def __init__(self, app_config, model_res, options):
        self.app_config = app_config
        self.model_res = model_res
       
        self.app_name = options['app_name']
        self.author_name = options['author_name']
        self.author_email = options['author_email']

        self.client_str = ""
        self.model_names = [model.__name__ for model in self.app_config.get_models()]

        self.generate_client()

    def get_relational_keys(self, model):
        relational_keys = []
        for field in model._meta.fields:
            if isinstance(self._meta.get_field_by_name(field.name)[0], models.ForeignKey):
                relational_keys.append(field.name)
            if isinstance(self._meta.get_field_by_name(field.name)[0], models.ManyToManyKey):
                relational_keys.append(field.name)
            if isinstance(self._meta.get_field_by_name(field.name)[0], models.OneToOneKey):
                relational_keys.append(field.name)
        if not relational_keys:
            return None
        return relational_keys

    def generate_client_imports(self):
        self.client_str += CLIENT_FILE_HEADER.format(
            lara_app=self.app_name,
            lara_author=self.author_name,
            lara_author_email=self.author_email,
        )
        self.client_str += f"""

import logging
import asyncio

# import api
#import {self.app_name}.grpc.{self.app_name}_pb2 as {self.app_name}_pb2

from {self.app_name}.models import {', '.join(self.model_names)}\n\n"""

    def generate_client(self):
        self.generate_client_imports()

        grpc_handler_str = ""
        grpc_handler_create_str = ""
        foreing_keys = []

        for model in self.app_config.get_models():
            fields = [field.name for field in model._meta.fields if "_id" not in field.name]
            relational_keys = self.get_relational_keys(model)
            # :TODO: use set compreshension
            non_rel_fields = fields - relational_keys
            #fields_param_str = ", ".join([f"{field}=None" for field in fields])
            fields_str = " ,".join([f"\n{4 * INDENT_WIDTH * ' '}{field}=" for field in fields])
            #fields_str = ",".join([f"\n{4 * INDENT_WIDTH * ' '}'{field}'" for field in fields])
            #fields_str = ", ".join([f"{field}'" for field in fields])
            

            grpc_handler_str += f"""{ model.__name__.lower()}_handler = {model.__name__.capitalize()}GRPCHandler(channel)\n"""
            grpc_handler_create_str += f"""{ model.__name__.lower()}_handler.create_{model.__name__.lower()}(n=1)\n"""

            self.client_str += f"""class {model.__name__.capitalize()}GRPCHandler():
    def __init__(self, channel):
        
        # create a client (stub)
        self.{model.__name__.lower()}_client = {self.app_name}_pb2.{self.app_name}Stub(channel)
         
    def create_{model.__name__.lower()}_entries(self, n=1):

        for i in range(n):
            res = self.{model.__name__.lower()}_client.Create({self.app_name}_pb2.{model.__name__.capitalize()}Request({fields_str}))
        \n\n
        
        # proto_class = {self.app_name}_pb2.{model.__name__.capitalize()}Response \n
        # proto_class_list = {self.app_name}_pb2.{model.__name__.capitalize()}ListResponse \n\n"""


        # main section

        self.client_str += f"""
async def main():
    # create a channel to a grpc server
    grpc_server_host = 'localhost' # '127.0.0.1' 
    grpc_server_port = 50066

    channel = grpc.aio.insecure_channel(f'{{grpc_server_host}}:{{grpc_server_port}}')

    {grpc_handler_str}\n
    # calling the create method of the stub for the given model
    {grpc_handler_create_str}\n
    

if __name__ == "__main__":
    asyncio.run(main())
"""

        client_filename = f"{self.app_name}_client.py"

        print(f"writing {client_filename} to {os.getcwd()}")
        print(self.client_str)

        # if os.path.isfile(client_filename):
        #     append = input(f"{client_filename} already exists, append to file? (y/n) ")
        #     if append.lower() == "y":
        #         with open(client, "a") as f:
        #             f.write(self.client_str)
        # else:
        #     # write sef.client_str to file
        #     with open(client, "w") as f:
        #         f.write(self.client_str)



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_base models *

:details: lara_django_base database models.
         -
         - fixtures can be generated with
         lara-django-dev dumpdata lara_django_base.itemstatus --indent 2 --natural-primary > 2_itemstatus_fix.json

         model_curi are compact URIs (CURIs) for the model, e.g. lara:base/ItemStatus.
         They are used for semantic representation.

:authors: mark doerr <mark.doerr@uni-greifswald.de>
          David Lacko @deiwo
          Jan Velecký

.. note:: -
.. todo:: -
________________________________________________________________________
"""

import datetime
import hashlib
import json
import uuid
from typing import Any
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.core.validators import RegexValidator
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from mptt.models import MPTTModel, TreeForeignKey

from .models_base import Currency, ExtraData, Namespace

settings.FIXTURES += [
    "3_geolocation_fix",
    # "4_location_fix",
    "4_room_category_fix",
    "1_countries_fix",
    "3_country_subdiv_type_fix",
    # "4_country_subdiv_fix",
    "4_cities_fix",
]


class URNField(models.CharField):
    """URN field that accepts URLs that adhere the scheme for URNs  only."""

    default_validators = [
        RegexValidator(
            regex=r"^urn:[a-z0-9][a-z0-9-]{1,31}:([a-z0-9()+,-.:=@;$_!*\'%/?#]|%[0-9a-f][0-9a-f])*$",
            message="Please enter a valid URN, s. https://en.wikipedia.org/wiki/Uniform_Resource_Name",
            code="invalid_URN",
        )
    ]  # [URLValidator(schemes=['urn'])]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.max_length = 512



class GeoLocation(models.Model):
    """
    Generic geo location information class to store geo location data.

    Like Decimal Degrees of the location of, e.g. a building, a sample, etc. ...

     .. todo:: accept only one input format, add converters to common formats
    """

    model_curi = "lara:base/GeoLocation"
    geolocation_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(unique=True, help_text="Unique location name")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    coordinates_dd_lat = models.DecimalField(
        max_digits=32,
        decimal_places=24,
        help_text="in Decimal degrees (DD) - latitude, e.g., 54.0915068",
    )
    coordinates_dd_long = models.DecimalField(
        max_digits=32,
        decimal_places=24,
        help_text="in Decimal degrees (DD) - longitude, e.g., 13.4029247",
    )
    elevation = models.IntegerField(
        null=True, blank=True, help_text="elevation in m above sea level"
    )
    # this could be autogenerated upon saved...
    openstreetmap = models.URLField(
        blank=True,
        max_length=1024,
        help_text="OpenStreetMap (www.openstreetmap.org) link to address",
    )
    # this could be autogenerated upon saved...
    googlemap = models.URLField(
        blank=True, max_length=1024, help_text="google maps link to address"
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_data_extra_related_query",
        blank=True,
        help_text="this can be used to extend the geo location: like maps to location etc.",
    )
    datetime_last_modified = models.DateTimeField(
        null=True, blank=True, help_text="date and time when data was last modified"
    )
    description = models.TextField(
        blank=True, help_text="description of the geo location"
    )

    class Meta:
        db_table_comment = (
            "Generic geo location information class to store geo location data,"
            "like Decimal Degrees of the location of, e.g. a building, a sample, etc. ..."
        )

    def __str__(self):
        return self.name or ""  # self.openstreetmap

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/GeoLocation/"
                + self.name.replace(" ", "_").lower().strip()
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    # coordinate converters
    def to_dms(self):
        """converting DD to Degrees, minutes, and seconds (DMS)."""
        # return DD_to_DMS( self.coordinates_dd_lat, self.coordinates_dd_long)
        # coordinates_DMS =  models.TextField(blank=True, default="""0deg0'0.0"N 0deg0'0.0"E""",
        # help_text="""in Degrees, minutes, and seconds (DMS): 54deg5'29.434"N 13deg24'10.527"E""")
        # # might need a syntax checker

    def to_dmm(self):
        """converting DD to Degrees and decimal minutes (DMM)"""
        # return DD_to_DMM(self.coordinates_dd_lat, self.coordinates_dd_long)
        # coordinates_DMM =  models.TextField(blank=True, default="0 0.0, 0 0.0",
        #  help_text="in Degrees and decimal minutes (DMM): 41 24.2028, 2 10.4418")


class Country(models.Model):
    """generic country representation class"""

    model_curi = "lara:base/Country"
    country_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(unique=True,
                            help_text="international name of the country")
    name_local = models.TextField(
        unique=True,
        blank=True, help_text="local name of the country, e.g. Deutschland"
    )
    alpha2code = models.CharField(
        max_length=2,
        unique=True,
        blank=True,
        help_text="International 2-letter code, like DE, SE, DK ...",
    )
    alpha3code = models.CharField(
        max_length=3,
        unique=True,
        help_text="International 3-letter code, like DEU, USA, HEL ..."
    )
    numeric_code = models.IntegerField(
        unique=True,
        blank=True,
        null=True,
        help_text="International 3 number code, like 276 (for DEU)",
    )
    phone_code = models.IntegerField(
        blank=True,
        null=True,
        unique=True,
        help_text="International telephone code, like 49, 45, ...",
    )
    icon = models.TextField(
        blank=True, help_text="XML/SVG icon / symbol / flag of the country"
    )
    currency = models.ForeignKey(
        Currency,
        related_name="%(app_label)s_%(class)s_currencies_related",
        related_query_name="%(app_label)s_%(class)s_currencies_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="currency",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    timezone = models.IntegerField(
        blank=True, null=True, help_text="time zone offset from UTC, e.g. 5, -3 "
    )
    description = models.TextField(blank=True, help_text="description of the country")

    class Meta:
        verbose_name_plural = "Countries"
        db_table_comment = "Generic country representation class"

    def __str__(self):
        return self.name_local or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/Country/"
                + self.name.replace(" ", "_").lower().strip()
            )

        if self.country_id is None:
            if self.numeric_code is not None:
                self.country_id = uuid.uuid5(uuid.NAMESPACE_URL, self.numeric_code)
            elif self.alpha3code is not None and self.alpha3code != "":
                self.country_id = uuid.uuid5(uuid.NAMESPACE_URL, self.alpha3code)
            else:
                self.country_id = uuid.uuid4()

        super().save(*args, **kwargs)


class SubdivisionType(models.Model):
    """Country Subdivision Type: e.g. State, Province, Emirate, ..."""

    model_curi = "lara:base/SubdivisionType"
    subdivisiontype_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )  # models.AutoField( primary_key=True)  #
    name = models.TextField(
        unique=True,
        help_text="subdivision type, e.g. State, Province, Emirate, ... ",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True, help_text="description of subdivision type"
    )

    class Meta:
        db_table_comment = (
            "Country Subdivision Type: e.g. State, Province, Emirate, ..."
        )

    def __str__(self):
        return self.name or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/SubdivisionType/"
                + self.name.replace(" ", "_").lower().strip()
            )

        super().save(*args, **kwargs)


class CountrySubdivision(models.Model):
    """Generic country subdivision  to fit federal states, provinces, counties etc. into the same concept
    Not limited to US federal states - also Germany and many other countries have them, s. pypi iso3166 and pycountry
    """

    model_curi = "lara:base/CountrySubdivision"
    countrysubdivision_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(
        blank=True,
        help_text="international name of the subdivision/federal state, Hessia, Thuringia, Baveria, ...",
    )
    name_local = models.TextField(
        blank=True,
        help_text="local name of the state, e.g. Hessen, Thüringen",
    )
    code = models.CharField(
        max_length=16,
        blank=True,
        help_text="multi-letter code, like US-MA, DE-TH ,  ...",
    )
    alpha2code = models.CharField(
        max_length=3,
        blank=True,
        help_text="2-letter code, like MA, BY, DK ...",
    )
    subdivision_type = models.ForeignKey(
        SubdivisionType,
        related_name="%(app_label)s_%(class)s_subdivision_types_related",
        related_query_name="%(app_label)s_%(class)s_subdivision_types_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    country = models.ForeignKey(
        Country,
        related_name="%(app_label)s_%(class)s_countries_related",
        related_query_name="%(app_label)s_%(class)s_countries_related_query",
        on_delete=models.CASCADE,
        help_text="country",
    )
    timezone = models.IntegerField(
        blank=True, null=True, help_text="time zone offset from UTC, e.g. 5, -3 "
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True, help_text="description of the country subdivision"
    )

    class Meta:
        db_table_comment = ("Generic country subdivision  to fit federal states,"
                            "provinces, counties etc. into the same concept.")
        constraints = [
            models.UniqueConstraint(
                fields=["name", "country"],
                name="%(app_label)s_%(class)s_name_country_unique",
            ),
        ]

    def __str__(self):
        if self.name_local != "":
            return f"{self.name} ({self.name_local})" or ""
        return self.name or ""

    def __repr__(self):
        if self.name_local != "":
            return f"{self.name} ({self.name_local})" or ""
        return self.name or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/CountrySubdivision/"
                + self.name.replace(" ", "_").lower().strip()
            )

        if self.timezone is None:
            self.timezone = self.country.timezone
        super().save(*args, **kwargs)


class City(models.Model):
    """Generic city representation class"""

    model_curi = "lara:base/City"
    city_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(
         help_text="international name of the city, e.g. Munich"
    )
    name_local = models.TextField(
        blank=True, help_text="local name of the city, e.g. München"
    )
    country = models.ForeignKey(
        Country,
        related_name="%(app_label)s_%(class)s_contries_related",
        related_query_name="%(app_label)s_%(class)s_countries_related_query",
        on_delete=models.CASCADE,
        null=True,
        help_text="country",
    )
    subdivision = models.ForeignKey(
        CountrySubdivision,
        related_name="%(app_label)s_%(class)s_subdivisions_related",
        related_query_name="%(app_label)s_%(class)s_subdivisions_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="country subdivision where the city is loacted",
    )
    timezone = models.IntegerField(
        blank=True, null=True, help_text="time zone offset from UTC, e.g. 5, -3 "
    )
    # ZIP codes (in extra table) ?
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(blank=True, help_text="description of the city")
    search_vector = SearchVectorField(null=True)

    class Meta:
        verbose_name_plural = "Cities"
        db_table_comment = "Generic city representation class"
        indexes = (GinIndex(fields=["search_vector"]),)
        constraints = [
            models.UniqueConstraint(
                fields=["name", "country"],
                name="%(app_label)s_%(class)s_name_country_unique",
            ),
        ]

    def __str__(self):
        return f"{self.name} ({self.name}/{self.country})" or ""

    def __repr__(self):
        return f"{self.name} ({self.name}/{self.country})" or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/City/"
                + self.name.replace(" ", "_").lower().strip()
            )

        if self.country is not None and self.timezone is None:
            self.timezone = self.country.timezone

        super().save(*args, **kwargs)


class Address(models.Model):
    """generic Address class:
    this class can be used to store all kind of Addresses"""

    model_curi = "lara:base/Address"
    address_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        help_text="namespace of address",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    building = models.TextField(
        blank=True,
        help_text="in case the institution is within a bigger complex/campus",
    )
    street = models.TextField(blank=True, help_text="street, excluding house number")
    house_number = models.TextField(blank=True, help_text="house number")
    supplement = models.TextField(
        blank=True,
        help_text="address supplement info to make it unique, like c/o, room/floor",
    )
    city = models.ForeignKey(
        City,
        related_name="%(app_label)s_%(class)s_cities_related",
        related_query_name="%(app_label)s_%(class)s_cities_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="city",
    )
    subdivision = models.ForeignKey(
        CountrySubdivision,
        related_name="%(app_label)s_%(class)s_subdivisions_related",
        related_query_name="%(app_label)s_%(class)s_subdivisions_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="e.g. federal state, like Thuringia or Alaska",
    )
    country = models.ForeignKey(
        Country,
        related_name="%(app_label)s_%(class)s_countries_related",
        related_query_name="%(app_label)s_%(class)s_countries_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="country",
    )
    postal_code = models.TextField(blank=True, help_text="postal code / ZIP")
    geolocation = models.ForeignKey(
        GeoLocation,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="Geographical location",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_data_extra_related_query",
        blank=True,
        help_text="extra information to addresses",
    )
    datetime_last_modified = models.DateTimeField(
        null=True, blank=True, help_text="date and time when data was last modified"
    )
    # iri = models.URLField(
    #     blank=True,  null=True, unique=True, max_length=512,
    #   help_text="International Resource Identifier - IRI: is used for semantic representation ")
    description = models.TextField(blank=True, help_text="description of the address")
    search_vector = SearchVectorField(null=True)

    class Meta:
        verbose_name_plural = "Addresses"
        db_table_comment = "Generic Address class: this class can be used to store all kind of Addresses"
        indexes = (
            GinIndex(fields=["search_vector"]),
            models.Index(
                fields=["namespace", "street", "postal_code", "city"],
                name="namespace_str_post_city_idx",
            ),
        )
        constraints = [
            models.UniqueConstraint(
                fields=["street", "house_number", "building", "postal_code", "city"],
                name="%(app_label)s_%(class)s_street_house_number_building_postal_code_city_unique",
            ),
        ]

    def __str__(self):
        house_number = self.house_number or ""
        postal_code = self.postal_code or ""
        alpha2code = self.country.alpha2code + "-" if self.country else ""
        return (
            f"{self.street} {house_number}, {alpha2code}{postal_code}, {self.city}"
            or ""
        )

    def __repr__(self):
        house_number = self.house_number or ""
        building = self.building or ""

        alpha2code = self.country.alpha2code + "-" if self.country else ""
        return (
            f"{self.street}, {house_number}, ({building}) {alpha2code}{self.postal_code}, {self.city}"
            or ""
        )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """

        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"] + "base/Address/" + str(self.address_id)
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    @property
    def name(self):
        """Returns a concatenation of street, building, house_number, postal_code, city and country"""
        building_info = f"({self.building})/" if self.building else ""
        number_info = self.house_number or ""
        postal_code = self.postal_code or ""
        city = self.city.name or ""
        country = self.country.name or ""

        return f"{building_info}{self.street}/{number_info}/{postal_code}/{city}/{country}".strip()


class RoomCategory(models.Model):
    """room category: office, lab, lab_S1, lab_S2, ... seminar room"""

    model_curi = "lara:base/RoomCategory"
    roomcategory_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    category = models.TextField(unique=True, help_text="room category")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(blank=True, help_text="room category description")

    class Meta:
        db_table_comment = (
            "Room category: office, lab, lab_S1, lab_S2, ... seminar room"
        )
        verbose_name_plural = "RoomCategories"

    def __str__(self):
        return self.category or ""

    def __repr__(self):
        return self.category or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/RoomCategory/"
                + self.category.replace(" ", "_").lower().strip()
            )

        super().save(*args, **kwargs)


class Room(models.Model):
    """Rooms and laboratories"""

    model_curi = "lara:base/Room"
    room_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of room",
    )

    name = models.TextField(
        blank=True, help_text="name of the room - some rooms have names"
    )
    category = models.ForeignKey(
        RoomCategory,
        related_name="%(app_label)s_%(class)s_categories_related",
        related_query_name="%(app_label)s_%(class)s_categories_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="room category",
    )
    number = models.TextField(blank=True, help_text="room number")
    floor = models.TextField(blank=True, help_text="Floor")
    phone_number = models.TextField(
        blank=True,
        help_text="room phone number, if possible add the +Country code notation, e.g., +44 3834 420 4411",
    )
    # address = models.ForeignKey(
    #     Address,
    #     related_name="%(app_label)s_%(class)s_addresses_related",
    #     related_query_name="%(app_label)s_%(class)s_addresses_related_query",
    #     on_delete=models.CASCADE,
    #     blank=True,
    #     null=True,
    #     help_text="reference to an address",
    # )  # check, if it can be 0 !
    room_map = models.ImageField(
        upload_to="people/rooms/",
        blank=True,
        null=True,
        help_text="location room map rel. path/filename to image",
    )
    # further info about the room
    max_people = models.IntegerField(
        null=True, blank=True, help_text="max number of people for this room"
    )
    area = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        blank=True,
        null=True,
        help_text="room area in square meters",
    )
    volume = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        blank=True,
        null=True,
        help_text="room volume in cubic meters",
    )
    features = models.JSONField(
        blank=True,
        null=True,
        help_text="features of this room, like sockets, lights, gas, furniture, infrastructure,...",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_room_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_room_data_extra_related_query",
        blank=True,
        help_text="extra information for this room",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(blank=True, help_text="room description")

    datetime_last_modified = models.DateTimeField(
        null=True, blank=True, help_text="date and time when data was last modified"
    )
    search_vector = SearchVectorField(null=True)

    class Meta:
        db_table_comment = "Rooms and laboratories"
        indexes = (GinIndex(fields=["search_vector"]),)
        constraints = [
            models.UniqueConstraint(
                fields=["namespace", "name", "number", "floor"],
                name="%(app_label)s_%(class)s_namespace_name_number_floor_unique",
            ),
        ]

    def __str__(self):
        return f"{self.name}" or ""

    def __repr__(self):
        return f"{self.name}" or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/Room/"
                + self.name.replace(" ", "_").lower().strip()
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class LocationType(models.Model):
    """Location type: e.g. shelf, drawer, box, container, device, ..."""

    model_curi = "lara:base/LocationType"
    locationtype_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(unique=True, help_text="location type name")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(blank=True, help_text="description of the location type")

    class Meta:
        db_table_comment = "Location type: e.g. shelf, drawer, box, container, device, ..."
        verbose_name_plural = "LocationTypes"

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

class Location(MPTTModel):
    """Physical location of objects like device, substances, organisms, containers etc.,
    at concrete place, e.g., in shelf or drawer in a lab or room.
    .. todo: a foreign key to a device for modelling a location in a device could be considered,
             but generates an additional dependency.....
    """

    model_curi = "lara:base/Location"
    location_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of location",
    )
    # name_display = models.TextField(
    #     help_text="display name of the location - some locations have display names"
    #     )
    name = models.TextField(unique=True, help_text="Unique location name")
    location_type = models.ForeignKey(
        LocationType,
        related_name="%(app_label)s_%(class)s_location_types_related",
        related_query_name="%(app_label)s_%(class)s_location_types_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="location type, e.g. shelf, drawer, box, container, device, ...",
    )
    parent = TreeForeignKey(
        "self",
        null=True,
        blank=True,
        related_name="%(app_label)s_%(class)s_experiment_parent_related",
        related_query_name="%(app_label)s_%(class)s_parent_related_query",
        on_delete=models.CASCADE,
        help_text="mptt - parent projects",
    )
    barcode = models.TextField(
        blank=True,
        unique=True,
        help_text="text representation of a barcode of the location.",
    )
    barcode_json = models.JSONField(
        blank=True,
        null=True,
        help_text="Advanced barcodes in JSON for multiple barcode representations, like QR codes etc.",
    )
    geolocation = models.ForeignKey(
        GeoLocation,
        related_name="%(app_label)s_%(class)s_geolocations_related",
        related_query_name="%(app_label)s_%(class)s_geolocations_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="geo location",
    )
    address = models.ForeignKey(
        Address,
        related_name="%(app_label)s_%(class)s_addresses_related",
        related_query_name="%(app_label)s_%(class)s_addresses_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="reference to an address",
    )
    location_json = models.JSONField(
        blank=True,
        null=True,
        help_text="JSON field for advanced location specification - can be used to store, e.g. locations/"
                  "positions in a device",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    # add PAC-ID
    url = models.URLField(
        blank=True,
        help_text="Universal Resource Locator - this can be used to link this location to an external location",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_data_extra_related_query",
        blank=True,
        help_text="like maps to location etc.",
    )
    datetime_last_modified = models.DateTimeField(
        null=True, blank=True, help_text="date and time when data was last modified"
    )
    description = models.TextField(blank=True, help_text="location description")
    search_vector = SearchVectorField(null=True)

    class Meta:
        db_table_comment = (
            "Physical location of objects like device, substances, organisms, containers"
            "etc., at concrete place, e.g., in shelf or drawer in a lab or room."
        )
        indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return f"{self.name} {self.description}" or ""

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for the IRI
        """
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/Location/"
                + self.name.replace(" ", "_").lower().strip()
            )

        if self.barcode == "" or self.barcode is None:
            self.barcode = f"AUTO_{str(uuid.uuid4())[:8]}"

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)




#!/usr/bin/env python3
"""
________________________________________________________________________.

:PROJECT: simple_installer

*A simple program installation library*

:details: This is a simple program installation library 
          including asking user descisions:
         - yes / no queries
         - text queries
         - simple help
         - default, predefined answers
         
         It also contains some convenient functions for 
         calling of system commands :
         - simple system command calling module
         - running setup.py
         - import and install if not available

:file:    simpleinstall.py

:author:  mark doerr (mark@ismeralda.org) :

:date: (creation)          20180310
:date: (last modification) 20180612

.. note::  to debug the setup.py installation, execute: 
           export DISTUTILS_DEBUG=TRUE
            
.. todo:: - OS/platform handling

________________________________________________________________________

**Copyright**:
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
  INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  For further Information see COPYING file that comes with this distribution.
________________________________________________________________________
"""

import os
import sys
import logging

import subprocess
import logging

from distutils.core import run_setup
from distutils.util import strtobool

__version__ = "0.0.1"

def query_yes_no(question, default_answer="yes", help=""):
    """Ask user at stdin a yes or no question
    
    :param question: question text to user
    :param default_answer: should be "yes" or "no"
    :param help: help text string
    :return:  :type: bool
    """
    if default_answer == "yes":
        prompt_txt = "{question} [Y/n] ".format(question=question)
    elif default_answer == "no" : # explicit no
        prompt_txt = "{question} [y/N] ".format(question=question)
    else:
        raise ValueError("default_answer must be 'yes' or 'no'!")

    while True:
        try:
            answer = input(prompt_txt)
            if answer:
                if answer=="?":
                    print(help)
                    continue
                else:
                    return strtobool(answer)
            else :
                return strtobool(default_answer)
        except ValueError:
            sys.stderr.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")
        except KeyboardInterrupt:
            logging.error("Query interrupted by user, exiting now ...")
            exit(0)

def query(question, default_answer="", help=""):
    """Ask user a question
    
    :param question: question text to user
    :param default_answer: any default answering text string
    :param help:  help text string
    :return: stripped answer string
    """
    prompt_txt = "{question} [{default_answer}] ".format(question=question, default_answer=default_answer)

    while True:
        answer = input(prompt_txt).strip()
        
        if answer :
            if answer=="?":
                print(help)
                continue
            else :
                return answer
        else :
            return default_answer


def call(command=""):
    ''' Convenient command call: it splits the command string into tokens (default separator: space)
    
    :param command: the command to be executed by the system
    '''
    try:
        cmd_lst = command.split()
        
        subprocess.run(cmd_lst, check=True)
    except subprocess.CalledProcessError as err:
        logging.error('ERROR:', err)
        
        
def run(command="", parameters=[]):
    '''This version is closer to the subprocess version
    
    :param command: the command to be executed by the system
    :param parameters: parameters of the this command
    '''
    try:
        subprocess.run([command]+parameters, check=True, shell=True)  #stdout=subprocess.PIPE
    except subprocess.CalledProcessError as err:
        logging.error('ERROR:', err)

def runSetup(src_dir="", lib_dir=""):
    """running a setup.py file within a pyhton script 
       it requires a lot of things set...
       :param src_dir: directory containing the setup.py file
       :param lib_dir: directory containing the target lib directory
    """
    # all path settings seem to be required by run_setup and setup.py
    os.environ["PYTHONPATH"] = os.path.join(lib_dir, 'lib', 'python3.5', 'site-packages')
    sys.path.append(lib_dir)
    os.chdir(src_dir)
    setup_file = os.path.join(src_dir, 'setup.py')
    run_setup(setup_file,  script_args=['install', '--prefix', lib_dir ])
            
def installAndImport(package):
    """ This imports a package and 
        if not available, installs it first 
    """
    import importlib
    try:
        importlib.import_module(package)
    except ImportError:
        import pip
        pip.main(['install', package])
    finally:
        globals()[package] = importlib.import_module(package)
        

if __name__=="__main__":
    """simple usage and testing routine """

    answer_txt = query("What do you like ?", default_answer="Spam", help="HELP: Here you can say, what you like")

    answer_bool = query_yes_no("Do you really like |{0}| ?".format(answer_txt), help="HELP: Here you can confirm, what you like")

    if answer_bool:
        sys.stdout.write("Indeed, she really likes |{0}| !\n".format(answer_txt))
    else:
        sys.stdout.write("Well, what a pitty. She does not like {0} !\n".format(answer_txt))

    if query_yes_no("Anything else ?", default_answer="no"):
        answer_txt = query("What else do you like ?", default_answer="Eggs")

        sys.stdout.write("I will also get some |{0}| for you !\n".format(answer_txt))
    else:
        sys.stdout.write("You are welcome !\n")




"""_____________________________________________________________________

:PROJECT: LARAsuite

*generate High Level Interface *

:details: generate_high_level_interface - a High Level Interface generator for LARA-django.
          This is a LARA-django-base management command to generate a High Level Interface for a
          given app. The generated interface can be used to generate various validators.
          s. https://linkml.io/linkml/index.html

          usage:
            lara-django-dev generate_high_level_interface <lara_app>

          once the interface is generated, it can be used to generate a pydantic model (from the LinkML package) with:
            gen-pydantic <lara_app>_interface.yaml

:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - async
          - delete: return no list.
________________________________________________________________________
"""

import logging
import os

from django.apps import apps
from django.conf import settings
from django.core.management.base import LabelCommand
from django.db import models

# Configurable constants
MAX_LINE_WIDTH = getattr(settings, "MAX_LINE_WIDTH", 120)
INDENT_WIDTH = getattr(settings, "INDENT_WIDTH", 2)

INTERFACE_FILE_HEADER = """\"\"\"_____________________________________________________________________

:PROJECT: LARAsuite

*{lara_app} high_level_interface *

:details: {lara_app} high_level_interface.
         - generated with 'lara-django-dev generate_high_level_interface {lara_app}
:authors: {lara_author} <{lara_author_email}>

.. note:: -
.. todo:: -
________________________________________________________________________
\"\"\""""


def indent(level=1, text="", end="\n"):
    return "\n".join([(level * INDENT_WIDTH * " ") + line for line in text.split("\n")]) + end


def lower_strip(s):
    return s.lower().replace(" ", "").strip()


class Command(LabelCommand):
    help = """Generate a High Level Interface for a given LARA app (models)"""
    args = "[app_name, author_name, author_email]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument("app_name", help="Name of the LARA app to generate the LinML Schema for")

        parser.add_argument("--author-name", help="Name of the author", default="lara")
        parser.add_argument("--author-email", help="Email of the author", default="lara@lara.org")
        parser.add_argument(
            "--force",
            action="store_true",
            help="overwrite existing files",
            default=False,
        )
        # parser.add_argument('model_name', nargs='*')

    # @signalcommand
    def handle(self, *args, **options):
        self.app_name = options["app_name"]

        logging.warning("!! only a scaffold is generated, please check/add content to the generated files !!!")

        try:
            app = apps.get_app_config(self.app_name)
        except LookupError:
            self.stderr.write("This command requires an existing app name as argument")
            self.stderr.write("Available apps:")
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write("    %s" % label)
            return

        model_res = []
        # for arg in options['model_name']:
        #     model_res.append(re.compile(arg, re.IGNORECASE))

        HigLevelInterfaceApp(app, model_res, options)


class HigLevelInterfaceApp:
    def __init__(self, app_config, model_res, options):
        self.app_config = app_config
        self.model_res = model_res

        self.app_name = options["app_name"]
        self.base_app_name = self.app_name[len("lara_django_") :]  # self.app_name.split("_")[2]
        self.force_overwrite = options["force"]
        self.author_name = options["author_name"]
        self.author_email = options["author_email"]

        self.interface_str = ""
        self.model_names = [model.__name__ for model in self.app_config.get_models()]

        self.generate_interface()

    def generate_interface(self):
        self.generate_interface_header()
        self.generate_interface_imports()
        self.generate_classes()

        # saving interface to file
        interface_filename = f"{self.app_name}_interface.py"

        print(f"writing {interface_filename} to {os.getcwd()}")
        # print(self.interface_str)

        if os.path.isfile(interface_filename):
            if self.force_overwrite:
                with open(interface_filename, "w") as f:
                    f.write(self.interface_str)
            else:
                append = input(
                    f"{interface_filename} already exists,\n press 'a' to append,\n press 'o' to overwrite,\n or 'n' to do nothing: "  # noqa: E501
                )
                if append.lower() == "a":
                    with open(interface_filename, "a") as f:
                        f.write(self.interface_str)
                elif append.lower() == "o":
                    with open(interface_filename, "w") as f:
                        f.write(self.interface_str)
        else:
            with open(interface_filename, "w") as f:
                f.write(self.interface_str)

    def generate_interface_header(self):
        self.interface_str += INTERFACE_FILE_HEADER.format(
            lara_app=self.app_name,
            lara_author=self.author_name,
            lara_author_email=self.author_email,
        )

    def generate_interface_imports(self):
        self.interface_str += f"""

import json
import asyncio
import logging
import grpc

from google.protobuf.struct_pb2 import Struct
from google.protobuf.json_format import MessageToDict

import lara_django_base_grpc.v1.lara_django_base_pb2 as lara_django_base_pb2
import lara_django_base_grpc.v1.lara_django_base_pb2_grpc as lara_django_base_pb2_grpc

import {self.app_name}_grpc.v1.{self.app_name}_pb2 as {self.app_name}_pb2
import {self.app_name}_grpc.v1.{self.app_name}_pb2_grpc as {self.app_name}_pb2_grpc

import  lara_django_base_grpc.lara_django_base_data_model as base_dm
import  lara_django_base_grpc.lara_django_base_interface as base_if
from  lara_django_base_grpc.abstract_interface import LARAInterface, GRPCChannelSingleton, import_model_instance_from_file, export_model_instance_to_file, upload_file, download_file, update_file, upload_image, update_image, download_image, id_cache

import  {self.app_name}_grpc.{self.app_name}_data_model as {self.base_app_name}_dm

logger = logging.getLogger(__name__)

\n\n"""

    def generate_classes(self):
        for model in self.app_config.get_models():
            has_namespace = False
            has_image = False
            has_file = False

            field_names = [field.name for field in model._meta.get_fields()]

            for field in model._meta.get_fields():
                match field:
                    case models.fields.files.FileField():
                        # ImageField is a subclass of FileField, determine, if it is an ImageField
                        if field.__class__.__name__ == "ImageField":
                            has_image = True
                        else:
                            has_file = True

                if field.name == "namespace":
                    has_namespace = True

            self.interface_str += (
                self.generate_class_init(model, has_namespace=has_namespace, has_file=has_file, has_image=has_image)
                + self.generate_class_create(model, has_namespace=has_namespace, has_file=has_file, has_image=has_image)
                + self.generate_class_retrieve(
                    model, has_namespace=has_namespace, has_file=has_file, has_image=has_image
                )
                + self.generate_class_list(model, has_namespace=has_namespace, has_file=has_file, has_image=has_image)
                + self.generate_class_update(model, has_namespace=has_namespace, has_file=has_file, has_image=has_image)
                + self.generate_class_delete(model, has_namespace=has_namespace, has_file=has_file, has_image=has_image)
            )

    def generate_class_init(self, model, has_namespace=False, has_file=False, has_image=False):
        model_name = model.__name__
        model_name_low = model.__name__.lower()

        file_image_id_stub = f"""
        self.item_id = None
        self.update_item_id = "{model_name_low}_id"
        self.stub = self.{model_name_low}_client
"""

        file_chunk = f"""
        self.file_download_info = {self.app_name}_pb2.FileDownloadInfo
        self.file_upload_chunk = {self.app_name}_pb2.FileUploadChunk
"""

        image_chunk = f"""
        self.image_upload_chunk = {self.app_name}_pb2.ImageUploadChunk
"""

        if has_namespace:
            return f"""#_________________  {model_name}  ______________________\n\n
class {model_name}Interface:
    \"\"\"Initialize the {model_name}Interface.\"\"\"

    def __init__(self, channel: grpc.Channel = None, 
                       default_namespace_uri : str = None,
                       default_namespace_id : str = None,
                       namespace_if = None ) -> None:
        self._default_namespace_uri = default_namespace_uri
        self._default_namespace_id = default_namespace_id

        if channel is None:
            cs = GRPCChannelSingleton()
            channel = cs.channel

        self.namespace_if = namespace_if if namespace_if is not None else base_if.NamespaceInterface(channel)

        # self.{model_name_low}_class_client = (
        #     {self.app_name}_pb2_grpc.{model_name}classByIRIControllerStub(channel)
        # )

        self.{model_name_low}_client = {self.app_name}_pb2_grpc.{model_name}ControllerStub(channel)
        {file_image_id_stub if has_file or has_image else ""}{file_chunk if has_file else ""}{image_chunk if has_image else ""}

        # self.{model_name_low}_full_client = {self.app_name}_pb2_grpc.{model_name}FullControllerStub(
        #     channel
        # )

    @property
    def default_namespace_uri(self):
        return self._default_namespace_uri

    @property
    def default_namespace_id(self):
        return self._default_namespace_id

    @classmethod
    async def build_if(cls, channel: grpc.Channel = None, default_namespace_uri: str = None):

        if channel is None:
            cs = GRPCChannelSingleton()
            channel = cs.channel

        namespace_if = base_if.NamespaceInterface(channel)
        try:
            default_namespace_id = await namespace_if.retrieve_id(namespace_uri=default_namespace_uri)
        except Exception as e:
            logger.error(f"Error retrieving namespace_id: {{e}}")

        return cls(channel=channel, default_namespace_uri=default_namespace_uri,  
                   default_namespace_id=default_namespace_id, namespace_if=namespace_if)

    async def set_default_namespace_uri(self, namespace_uri: str = None):
        if namespace_uri is not None:
            self._default_namespace_uri = namespace_uri.strip()
        try:
            self._default_namespace_id = await self.namespace_if.retrieve_id(namespace_uri=namespace_uri)
        except Exception as e:
            logger.error(f"Error retrieving namespace_id: {{e}}")
    """
        ###  no namespace
        return f"""#_________________  {model_name}  ______________________\n\n
class {model_name}Interface:
    \"\"\"Initialize the {model_name}Interface.\"\"\"

    def __init__(self, channel: grpc.Channel = None) -> None:
        if channel is None:
            cs = GRPCChannelSingleton()
            channel = cs.channel

        self.{model_name_low}_client = {self.app_name}_pb2_grpc.{model_name}ControllerStub(channel)
        {file_image_id_stub if has_file or has_image else ""}{file_chunk if has_file else ""}{image_chunk if has_image else ""}
    """

    # replace variable names by shortcuts defined above
    def generate_class_create(self, model, has_namespace=False, has_file=False, has_image=False):
        model_name = model.__name__
        model_name_low = model.__name__.lower()
        # model_verbose_name = model._meta.verbose_name
        model_verbose_name_plural = lower_strip(model._meta.verbose_name_plural)

        upload_file_decorator = "\n    @upload_file  # needed for file upload"
        upload_image_decorator = "\n    @upload_image  # needed for image upload"

        model_class_id = f"{model_name_low}class_id = None"
        if has_namespace:
            return f"""
    @import_model_instance_from_file(){upload_file_decorator if has_file else ""}{upload_image_decorator if has_image else ""}
    async def create(self,  {lower_strip(model_verbose_name_plural)}:  list[{self.base_app_name}_dm.{model_name}] | None = None,
                            ids_only: bool = False,
                            namespace_uri: str = None) -> list[str] | list[{self.base_app_name}_dm.{model_name}]:
        \"\"\"
        Create a list of {model_name_low} instances.

            Returns a list of {model_name_low} data model objects or ids_only.

        :param {lower_strip(model_verbose_name_plural)}: list of {model_name_low} data model objects
        :param ids_only: return only the ids of the created instances
        :return: list of {model_name_low} data model objects or ids_only
        \"\"\"
        {model_class_id if has_namespace else ""}
        namespace_id = None
        if namespace_uri is None and self._default_namespace_uri is not None:
            namespace_uri = self._default_namespace_uri
            namespace_id = self._default_namespace_id
        if namespace_uri is None:
            raise ValueError("No namespace URI is given.")
        if namespace_id is None:
            try:
                namespace_id = await self.namespace_if.retrieve_id(namespace_uri=namespace_uri)
            except Exception as e:
                logger.error(f"Error retrieving namespace_id: {{e}}")

        # some sample code for retrieving a class IRI and adding namespace and class to the model
        # if {model_name_low}_class_iri is None:
        #     raise ValueError("No sample class IRI is given.")
        # try:
        #     res = await self.{model_name_low}_class_client.Retrieve(
        #         {self.app_name}_pb2.{model_name}classRetrieveRequest(iri={model_name_low}_class_iri)
        #     )
        #     {model_name_low}class_id = res.{model_name_low}class_id
        # except Exception as e:
        #     logger.error(f"Error retrieving {model_name_low}class_id: {{e}}")
        for {model_name_low} in {model_verbose_name_plural}:
            if {model_name_low}.namespace == "" or {model_name_low}.namespace == "default":
                    {model_name_low}.namespace = namespace_id
            # {model_name_low}.{model_name_low}_class = {model_name_low}class_id # uncomment if a model class is used

        try:
            create_coroutines = ( self.{model_name_low}_client.Create({self.app_name}_pb2.{model_name}Request(**{model_name_low}.model_dump())) for {model_name_low} in {model_verbose_name_plural} )
            res = await asyncio.gather(*create_coroutines)
            if ids_only:
                return [ MessageToDict(item, preserving_proto_field_name=True).get("{model_name_low}_id") for item in res ]
            else:
                return res

        except Exception as e:
            logger.error(f"Error creating {model_name_low}: {{e}}")
        """
        ###  no namespace
        return f"""
    @import_model_instance_from_file(){upload_file_decorator if has_file else ""}{upload_image_decorator if has_image else ""}
    async def create(self,  {model_verbose_name_plural}:  list[{self.base_app_name}_dm.{model_name}] | None = None,  ids_only: bool = False) -> list[str] | list[{self.base_app_name}_dm.{model_name}]:
        \"\"\"
        Create a list of {model_name_low} instances.

            Returns a list of {model_name_low} data model objects or ids_only.
        :param {model_verbose_name_plural}: list of {model_name_low} data model objects
        :param ids_only: return only the ids of the created instances
        :return: list of {model_name_low} data model objects or ids_only
        \"\"\"
        try:
            create_coroutines = ( self.{model_name_low}_client.Create({self.app_name}_pb2.{model_name}Request(**{model_name_low}.model_dump())) for {model_name_low} in {model_verbose_name_plural} )
            res = await asyncio.gather(*create_coroutines)
            if ids_only:
                return [item.{model_name_low}_id for item in res]
            else:
                return res

        except Exception as e:
            logger.error(f"Error creating {model_name_low}: {{e}}")
        """

    def generate_class_retrieve(self, model, has_namespace=False, has_file=False, has_image=False):
        model_name = model.__name__
        model_name_low = model.__name__.lower()

        download_file_decorator = "\n    @download_file  # needed for file download"
        download_image_decorator = "\n    @download_image  # needed for image download"

        get_file_method = f"""
    @download_file  # needed for file download
    async def get_file(self, {model_name_low}_id : str | None = None, id_only : bool = True ) -> bytes | None:
        \"\"\"Retrieve a file by data_id.\"\"\"
        self.item_id = {model_name_low}_id
        """

        get_file_by_name_namespace_method = f"""
    async def get_file_by_name(self, name: str | None = None,
                                namespace_uri : str | None = None,
                                filename_target : str | None = None,
                                as_byte_str : bool | None = None ) -> bytes | None:
        \"\"\"Retrieve a file by name.\"\"\"
        if namespace_uri is None and self._default_namespace_uri is not None:
            namespace_uri = self._default_namespace_uri
            # logger.warning(f"no namespace_uri provided, using default namespace: {{namespace_uri}}")
        if namespace_uri is None:
            raise ValueError("No namespace URI is given.")
        try:
            {model_name_low}_id = await self.get_by_name(name=name, namespace_uri=namespace_uri, id_only=True)
            return await self.get_file({model_name_low}_id={model_name_low}_id, filename_target=filename_target, as_byte_str=as_byte_str, get_file=True)
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low} name: {{e}}")
        """  # noqa: E501

        get_file_by_name_method = f"""
    async def get_file_by_name(self, name: str | None = None,
                                filename_target : str | None = None,
                                as_byte_str : bool | None = None ) -> bytes | None:
        \"\"\"Retrieve a file by name.\"\"\"
        try:
            {model_name_low}_id = await self.get_by_name(name=name, id_only=True)
            return await self.get_file({model_name_low}_id={model_name_low}_id, filename_target=filename_target, as_byte_str=as_byte_str,  get_file=True)
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low} name: {{e}}")
        """  # noqa: E501
        get_image_method = f"""
    @download_image  # needed for image download
    async def get_image(self, {model_name_low}_id : str = None, id_only : bool = True) -> bytes | None:
        \"\"\"Retrieve an image by data_id.\"\"\"
        self.item_id = {model_name_low}_id
        """

        if has_namespace:
            return f"""
    @export_model_instance_to_file
    async def retrieve(
        self,
        namespace_uri: str | None = None,
        {model_name_low}_id: str | None = None,
        name: str | None = None,
        # name_display: str | None = None,
        filter_dict: dict | None = None,
        raw : bool = False
        ) -> list[{self.base_app_name}_dm.{model_name}]:
        \"\"\"
        Retrieve a list of {model_name_low} instances by {model_name_low}_id, name or filter_dict.

            Returns a list of {model_name_low} data model objects.
                If raw is True, the raw protobuf objects are returned.
                If namespace_uri is "default", the default namespace_uri is used.
        \"\"\"
        if namespace_uri == "default":
            if self._default_namespace_uri is not None:
                namespace_uri = self._default_namespace_uri
            else:
                raise ValueError("No default namespace URI is given.")

        results_list = []
        if  {model_name_low}_id is not None:
            try:
                res =  await self.{model_name_low}_client.Retrieve(
                    {self.app_name}_pb2.{model_name}RetrieveRequest({model_name_low}_id={model_name_low}_id) )
                if raw:
                    results_list.append(res)
                else:
                    results_list.append( {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True)) )
            except Exception as e:
                logger.error(f"Error retrieving {model_name_low}_id: {{e}}")

        if filter_dict is None:
            filter_dict = {{}}

        if name is not None:
            filter_dict["name"] = name
        if namespace_uri is not None:
            filter_dict["namespace__uri"] = namespace_uri
        # if name_display is not None:
        #    filter_dict["name_display"] = name_display

        metadata = (("filters", (json.dumps(filter_dict))),)
        try:
            res = await self.{model_name_low}_client.List(
                    {self.app_name}_pb2.{model_name}ListRequest(), metadata=metadata
                    )
            if raw:
                    results_list += res.results
            else:
                # :TODO: a queue could be used here to speed up the retrieval
                results = res.results
                results_list += [ {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True)) for res in results ]
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low} name: {{e}}")

        return results_list

    #@id_cache
    async def retrieve_id(self,
                          name: str | None = None,
                          # name_display: str = None,
                          namespace_uri : str | None = None,
                          filter_dict : dict | None = None,
                          iri: str | None = None) -> str | None:
        \"\"\"Retrieve the {model_name_low}_id for a given name, name_display or iri.\"\"\"

        results_list = await self.retrieve(name=name, 
                        namespace_uri=namespace_uri, filter_dict=filter_dict, raw=True)

        if len(results_list) == 1 :
            return results_list[0].{model_name_low}_id
        else:
            raise ValueError(f"Error retrieving {model_name_low}_id - no or too many results found.")
    {download_file_decorator if has_file else ""}{download_image_decorator if has_image else ""}
    async def get_by_id(self, {model_name_low}_id: str = None, id_only: bool = False) -> {self.base_app_name}_dm.{model_name}:
        \"\"\"Retrieve a {model_name_low} data model by {model_name_low}_id.\"\"\"
        try:
            res = await self.{model_name_low}_client.Retrieve(
                {self.app_name}_pb2.{model_name}RetrieveRequest({model_name_low}_id={model_name_low}_id)
            )
            item = {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True))
            self.item_id = item.{model_name_low}_id
            return item
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low}_id: {{e}}")
    {download_file_decorator if has_file else ""}{download_image_decorator if has_image else ""}
    async def get_by_name(self, name: str = None,
                                namespace_uri : str = None,
                                id_only: bool = False) -> {self.base_app_name}_dm.{model_name} | str:
        \"\"\"Retrieve a {model_name_low} data model by name.\"\"\"
        if namespace_uri is None and self._default_namespace_uri is not None:
            namespace_uri = self._default_namespace_uri
        if namespace_uri is None:
            raise ValueError("No namespace URI is given.")
        try:
            res = await self.{model_name_low}_client.RetrieveByNameNamespace(
                {self.app_name}_pb2.{model_name}RetrieveByNameNSRequest(name=name, namespace_uri=namespace_uri)
            )
            item = MessageToDict(res, preserving_proto_field_name=True)
            self.item_id = item["{model_name_low}_id"]
            if id_only:
                return self.item_id
            else:
                return {self.base_app_name}_dm.{model_name}(**item)
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low} name: {{e}}")

    {get_file_method if has_file else ""}
    {get_file_by_name_namespace_method if has_file else ""}
    {get_image_method if has_image else ""}
    """  # noqa: E501
        ###  no namespace
        return f"""
    @export_model_instance_to_file
    async def retrieve(
        self,
        {model_name_low}_id: str | None = None,
        name: str | None = None,
        # name_display: str = None,
        filter_dict: dict | None = None,
        raw : bool = False
        ) -> list[{self.base_app_name}_dm.{model_name}]:
        \"\"\"
        Retrieve a list of {model_name_low} instances by {model_name_low}_id, name or filter_dict.

            Returns a list of {model_name_low} data model objects.
            If raw is True, the raw protobuf objects are returned.
        \"\"\"
        results_list = []
        if  {model_name_low}_id is not None:
            try:
                res =  await self.{model_name_low}_client.Retrieve(
                    {self.app_name}_pb2.{model_name}RetrieveRequest({model_name_low}_id={model_name_low}_id) )
                if raw:
                    results_list.append(res)
                else:
                    results_list.append( {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True)) )
            except Exception as e:
                logger.error(f"Error retrieving {model_name_low}_id: {{e}}")

        if filter_dict is None:
            filter_dict = {{}}

        if name is not None:
            filter_dict["name"] = name
        # if name_display is not None:
        #    filter_dict["name_display"] = name_display

        metadata = (("filters", (json.dumps(filter_dict))),)
        try:
            res = await self.{model_name_low}_client.List(
                    {self.app_name}_pb2.{model_name}ListRequest(), metadata=metadata
                    )
            if raw:
                    results_list += res.results
            else:
                # a queue could be used here to speed up the retrieval
                results = res.results
                results_list += [ {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True)) for res in results ]
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low} name: {{e}}")

        return results_list

    #@id_cache
    async def retrieve_id(self,
                          name: str | None = None,
                          # name_display: str = None,
                          filter_dict : dict | None = None,
                          iri: str | None = None) -> str | None:
        \"\"\"Retrieve the {model_name_low}_id for a given name, (name_display) or iri.\"\"\"
        results_list = await self.retrieve(name=name, filter_dict=filter_dict, raw=True)
        if len(results_list) == 1 :
            return results_list[0].{model_name_low}_id
        else:
            raise ValueError(f"Error retrieving {model_name_low}_id - no or too many results found.")
    {download_file_decorator if has_file else ""}{download_image_decorator if has_image else ""}
    async def get_by_id(self, {model_name_low}_id: str = None) -> {self.base_app_name}_dm.{model_name}:
        \"\"\"Retrieve a {model_name_low} data model by {model_name_low}_id.\"\"\"
        try:
            res = await self.{model_name_low}_client.Retrieve(
                {self.app_name}_pb2.{model_name}RetrieveRequest({model_name_low}_id={model_name_low}_id)
            )
            return {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True))
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low}_id: {{e}}")
    {download_file_decorator if has_file else ""}{download_image_decorator if has_image else ""}
    async def get_by_name(self, name: str = None, id_only: bool = False) -> {self.base_app_name}_dm.{model_name} | str:
        \"\"\"Retrieve a {model_name_low} data model by name.\"\"\"
        try:
            res = await self.{model_name_low}_client.RetrieveByNameNamespace(
                {self.app_name}_pb2.{model_name}RetrieveByNameNSRequest(name=name, namespace_uri="")
            )
            if id_only:
                return MessageToDict(res, preserving_proto_field_name=True)["{model_name_low}_id"]
            else:
                return {self.base_app_name}_dm.{model_name}(**MessageToDict(res, preserving_proto_field_name=True))
        except Exception as e:
            logger.error(f"Error retrieving {model_name_low} name: {{e}}")
    {get_file_method if has_file else ""}
    {get_file_by_name_method if has_file else ""}
    {get_image_method if has_image else ""}
    """

    def generate_class_list(self, model, has_namespace=False, has_file=False, has_image=False):
        model_name = model.__name__
        model_name_low = model.__name__.lower()

        return f"""
    async def list_items(self, filter_dict: dict | None = None, ids_only: bool = False) -> list[str]:
        \"\"\"List all {model_name_low}s.\"\"\"
        if filter_dict is None:
            res = await self.{model_name_low}_client.List({self.app_name}_pb2.{model_name}ListRequest())
        else:
            metadata = (("filters", (json.dumps(filter_dict))),)
            res = await self.{model_name_low}_client.List(
                {self.app_name}_pb2.{model_name}ListRequest(), metadata=metadata
            )
        if res is not None and ids_only:
            return [item.{model_name_low}_id for item in res.results]
        else:
            return res.results

    async def list_full(self, filter_dict: dict | None = None) -> list:
        \"\"\"List all {model_name_low}s.\"\"\"
        if self.{model_name_low}_full_client is None:
            raise ValueError("No full client available.")
        if filter_dict is None:
            return await self.{model_name_low}_full_client.List(
                {self.app_name}_pb2.{model_name}FullListRequest()
            )
        else:
            metadata = (("filters", (json.dumps(filter_dict))),)
            return await self.{model_name_low}_full_client.List(
                {self.app_name}_pb2.{model_name}FullListRequest(), metadata=metadata
            )
    """

    def generate_class_update(self, model, has_namespace=False, has_file=False, has_image=False):
        model_name = model.__name__
        model_name_low = model.__name__.lower()

        update_file_decorator = "\n    @update_file  # needed for file update"
        update_image_decorator = "\n    @update_image  # needed for image update"

        return f"""
    @import_model_instance_from_file(){update_file_decorator if has_file else ""}{update_image_decorator if has_image else ""}
    async def update(self, {model_name_low} : {self.base_app_name}_dm.{model_name} = None) -> None:
        \"\"\"Update a {model_name_low} model instance.\"\"\"
        try:
            res = await self.{model_name_low}_client.Update(
                {self.app_name}_pb2.{model_name}Request(**{model_name_low}.model_dump()
            ))
        except Exception as e:
            logger.error(f"Error updating {model_name_low}_id: {{e}}")
        else:
            return res
    """

    def generate_class_delete(self, model, has_namespace=False, has_file=False, has_image=False):
        model_name = model.__name__
        model_name_low = model.__name__.lower()

        return f"""
    async def delete(self, id_list: list[str] | None = None, filter_dict: dict | None = None) -> None:
        \"\"\"Delete a {model_name_low} by {model_name_low}_id or filter_dict.\"\"\"
        if filter_dict is not None:
            id_list = await self.list_items(filter_dict=filter_dict, ids_only=True)

        try:
            delete_coroutines = ( self.{model_name_low}_client.Destroy({self.app_name}_pb2.{model_name}DestroyRequest({model_name_low}_id=id)) for id in id_list )
            await asyncio.gather(*delete_coroutines)

        except Exception as e:
            logger.error(f"Error deleting {model_name_low}_id: {{e}}")
\n"""



"""_____________________________________________________________________

:PROJECT: lara-django

*django*

:details: django 
   core/management/base.py
   https://stackoverflow.com/questions/31838882/check-for-pending-django-migrations/49310733

:authors: mark doerr (mark@uni-greifswald.de)

:date: (creation)          20210416

.. note:: -
.. todo:: -
________________________________________________________________________

**Copyright**:
  This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
  INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  For further Information see LICENSE file that comes with this distribution.
________________________________________________________________________
"""

__version__ = "0.0.1"

import os
import logging

from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command
# ~ from django.core.management.commands import loaddata
from django.core.exceptions import ImproperlyConfigured
from django.core.management.color import color_style, no_style
from django.db import DEFAULT_DB_ALIAS, connections
from django.conf import settings


class Command(BaseCommand):
    """see https://docs.djangoproject.com/en/1.9/howto/custom-management-commands/ for more details
       using now new argparse mechanism of django > 1.8
    """
    # logging.basicConfig(format='%(levelname)s| %(module)s.%(funcName)s:%(message)s',
    #                    level=logging.DEBUG)  # level=logging.ERROR

    help = 'show migration information.'

    def add_arguments(self, parser):
        """ command line arguments s. https://docs.python.org/2/library/argparse.html#module-argparse """

        parser.add_argument('-n', '--num-pending-migrations',
                            action='store_true',
                            help='number of pending migrations')

        parser.add_argument('-l', '--list-pending-migrations',
                            action='store_true',
                            help='list of pending migrations')

        parser.add_argument('-c', '--check-migrations',
                            action='store_true',
                            help='check  migrations')
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help=(
                'Nominates a database to introspect. Defaults to using the "default" '
                "database."
            ),
        )
        
        parser.add_argument('-t', '--check-db-table',
                            action='store',
                            help='check for table in database')

    def handle(self, *args, **options):
        """Dispatcher based on commandline arguments/options"""

        if options['num_pending_migrations']:
            self.num_migrations()
        elif options['list_pending_migrations']:
            self.list_migrations()
        elif options['check_migrations']:
            self.check_migrations()
        elif options['check_db_table']:
            self.check_db_table(options)

    def num_migrations(self):
        """Initialises the LARA database
           :param options:
        """
        from django.db.migrations.executor import MigrationExecutor
        try:
            executor = MigrationExecutor(connections[DEFAULT_DB_ALIAS])
        except ImproperlyConfigured:
            # No databases are configured (or the dummy one)
            return

        plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
        if plan:
            print(len(plan))

    def list_migrations(self):
        """
        Print a warning if the set of migrations on disk don't match the
        migrations in the database.
        """
        from django.db.migrations.executor import MigrationExecutor
        try:
            executor = MigrationExecutor(connections[DEFAULT_DB_ALIAS])
        except ImproperlyConfigured:
            # No databases are configured (or the dummy one)
            return

        plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
        if plan:
            apps_waiting_migration = sorted(
                {migration.app_label for migration, backwards in plan})
            print(", ".join(apps_waiting_migration))

    def check_migrations(self):
        """
        Print a warning if the set of migrations on disk don't match the
        migrations in the database.
        """
        from django.db.migrations.executor import MigrationExecutor
        try:
            executor = MigrationExecutor(connections[DEFAULT_DB_ALIAS])
        except ImproperlyConfigured:
            # No databases are configured (or the dummy one)
            return

        plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
        if plan:
            apps_waiting_migration = sorted(
                {migration.app_label for migration, backwards in plan})
            self.stdout.write(
                self.style.NOTICE(
                    "\nYou have %(unapplied_migration_count)s unapplied migration(s). "
                    "Your project may not work properly until you apply the "
                    "migrations for app(s): %(apps_waiting_migration)s." % {
                        "unapplied_migration_count": len(plan),
                        "apps_waiting_migration": ", ".join(apps_waiting_migration),
                    }
                )
            )
            self.stdout.write(self.style.NOTICE(
                "Run 'python manage.py migrate' to apply them."))
            
    def check_db_table(self, options ):
        connection = connections[options["database"]]
        table_name = options["check_db_table"]
        return table_name in connection.introspection.table_names()



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_base admin *

:details: lara_django_base admin module admin backend configuration.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev tables_generator lara_django_base >> tables.py" to update this file
________________________________________________________________________
"""
# django Tests s. https://docs.djangoproject.com/en/4.1/topics/testing/overview/ for lara_django_base
# generated with django-extensions tests_generator  lara_django_base > tests.py (LARA-version)

import django_tables2 as tables

from .models import Address
from .models import City
from .models import Country
from .models import CountrySubdivision
from .models import Currency
from .models import DataType
from .models import ExtraData
from .models import GeoLocation
from .models import ItemStatus
from .models import Location
from .models import MediaType
from .models import Namespace
from .models import PhysicalUnit
from .models import Room
from .models import RoomCategory
from .models import ScientificConstant
from .models import SubdivisionType
from .models import Tag


class DataTypeTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:datatype-detail', [tables.A('pk')]))

    class Meta:
        model = DataType

        fields = ("name", "description")


class MediaTypeTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:mediatype-detail', [tables.A('pk')]))

    class Meta:
        model = MediaType

        fields = ("name", "media_type", "def_filename_extension_regex", "description")


class NamespaceTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:namespace-detail', [tables.A('pk')]))

    class Meta:
        model = Namespace

        fields = ("domain", "subdomain", "uri", "description")


class TagTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:tag-detail', [tables.A('pk')]))

    class Meta:
        model = Tag

        fields = ("tag",)


class ItemStatusTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:itemstatus-detail', [tables.A('pk')]))

    class Meta:
        model = ItemStatus

        fields = ("status", "description")


class ExtraDataTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:extradata-detail', [tables.A('pk')]))

    class Meta:
        model = ExtraData

        fields = (
            "name",
            "namespace",
            "uri",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
            "image",
        )


class PhysicalUnitTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:physicalunit-detail', [tables.A('pk')]))

    class Meta:
        model = PhysicalUnit

        fields = ("name", "dimension", "description")


class ScientificConstantTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:ScientificConstant-detail', [tables.A('pk')]))

    class Meta:
        model = ScientificConstant

        fields = (
            "name",
            "value",
            "unit",
            "iri",
            "url_source",
            "url_definition",
            "description",
        )


class CurrencyTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:currency-detail', [tables.A('pk')]))

    class Meta:
        model = Currency

        fields = (
            "name",
            "iri",
            "alpha3code",
            "numeric_code",
            "symbol",
            "exchange_rate_eur",
        )


class GeoLocationTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:geolocation-detail', [tables.A('pk')]))

    class Meta:
        model = GeoLocation

        fields = (
            "name",
            "coordinates_dd_lat",
            "coordinates_dd_long",
            "elevation",
            "openstreetmap",
            "googlemap",
        )


class CountryTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:country-detail', [tables.A('pk')]))

    class Meta:
        model = Country

        fields = (
            "name",
            "name_local",
            "alpha2code",
            "alpha3code",
            "numeric_code",
            "phone_code",
            "icon",
            "currency",
        )


class SubdivisionTypeTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:subdivisiontype-detail', [tables.A('pk')]))

    class Meta:
        model = SubdivisionType

        fields = ("name", "description")


class CountrySubdivisionTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:countrysubdivision-detail', [tables.A('pk')]))

    class Meta:
        model = CountrySubdivision

        fields = (
            "name",
            "name_local",
            "code",
            "alpha2code",
            "subdivision_type",
            "country",
            "timezone",
        )


class CityTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:city-detail', [tables.A('pk')]))

    class Meta:
        model = City

        fields = ("name", "name_local", "country", "subdivision", "timezone")


class AddressTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:address-detail', [tables.A('pk')]))

    class Meta:
        model = Address

        fields = (
            "building",
            "street",
            "house_number",
            "supplement",
            "city",
            "subdivision",
            "country",
            "postal_code",
            "geolocation",
        )


class RoomCategoryTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:roomcategory-detail', [tables.A('pk')]))

    class Meta:
        model = RoomCategory

        fields = ("category", "description")


class RoomTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:room-detail', [tables.A('pk')]))

    class Meta:
        model = Room

        fields = (
            "name",
            "category",
            "number",
            "floor",
            "phone_number",
            "room_map",
            "max_people",
            "area",
            "volume",
        )


class LocationTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_base:location-detail', [tables.A('pk')]))

    class Meta:
        model = Location

        fields = (
            "name",
            "barcode",
            "geolocation",
            "address",
            "building",
            "floor",
            "room",
            "cabinet",
            "shelf",
            "drawer",
            "box",
            "position",
            "uri",
            "iri",
            "url",
            "description",
        )
